#!/usr/bin/env python3
#
# This file is part of the MicroPython project, http://micropython.org/
#
# The MIT License (MIT)
#
# Copyright (c) 2014-2021 Damien P. George
# Copyright (c) 2017 Paul Sokolovsky
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

"""
pyboard interface

This module provides the Pyboard class, used to communicate with and
control a MicroPython device over a communication channel. Both real
boards and emulated devices (e.g. running in QEMU) are supported.
Various communication channels are supported, including a serial
connection, telnet-style network connection, external process
connection.

Example usage:

    import pyboard
    pyb = pyboard.Pyboard('/dev/ttyACM0')

Or:

    pyb = pyboard.Pyboard('192.168.1.1')

Then:

    pyb.enter_raw_repl()
    pyb.exec('import pyb')
    pyb.exec('pyb.LED(1).on()')
    pyb.exit_raw_repl()

Note: if using Python2 then pyb.exec must be written as pyb.exec_.
To run a script from the local machine on the board and print out the results:

    import pyboard
    pyboard.execfile('test.py', device='/dev/ttyACM0')

This script can also be run directly.  To execute a local script, use:

    ./pyboard.py test.py

Or:

    python pyboard.py test.py

"""

import ast
import errno
import os
import struct
import sys
import time

from collections import namedtuple

try:
    stdout = sys.stdout.buffer
except AttributeError:
    # Python2 doesn't have buffer attr
    stdout = sys.stdout


def stdout_write_bytes(b):
    b = b.replace(b"\x04", b"")
    stdout.write(b)
    stdout.flush()


class PyboardError(Exception):
    def convert(self, info):
        if len(self.args) >= 3:
            if b"OSError" in self.args[2] and b"ENOENT" in self.args[2]:
                return OSError(errno.ENOENT, info)

        return self


listdir_result = namedtuple("dir_result", ["name", "st_mode", "st_ino", "st_size"])


class TelnetToSerial:
    def __init__(self, ip, user, password, read_timeout=None):
        self.tn = None
        import telnetlib

        self.tn = telnetlib.Telnet(ip, timeout=15)
        self.read_timeout = read_timeout
        if b"Login as:" in self.tn.read_until(b"Login as:", timeout=read_timeout):
            self.tn.write(bytes(user, "ascii") + b"\r\n")

            if b"Password:" in self.tn.read_until(b"Password:", timeout=read_timeout):
                # needed because of internal implementation details of the telnet server
                time.sleep(0.2)
                self.tn.write(bytes(password, "ascii") + b"\r\n")

                if b"for more information." in self.tn.read_until(
                    b'Type "help()" for more information.', timeout=read_timeout
                ):
                    # login successful
                    from collections import deque

                    self.fifo = deque()
                    return

        raise PyboardError("Failed to establish a telnet connection with the board")

    def __del__(self):
        self.close()

    def close(self):
        if self.tn:
            self.tn.close()

    def read(self, size=1):
        while len(self.fifo) < size:
            timeout_count = 0
            data = self.tn.read_eager()
            if len(data):
                self.fifo.extend(data)
                timeout_count = 0
            else:
                time.sleep(0.25)
                if self.read_timeout is not None and timeout_count > 4 * self.read_timeout:
                    break
                timeout_count += 1

        data = b""
        while len(data) < size and len(self.fifo) > 0:
            data += bytes([self.fifo.popleft()])
        return data

    def write(self, data):
        self.tn.write(data)
        return len(data)

    def inWaiting(self):
        n_waiting = len(self.fifo)
        if not n_waiting:
            data = self.tn.read_eager()
            self.fifo.extend(data)
            return len(data)
        else:
            return n_waiting


class ProcessToSerial:
    "Execute a process and emulate serial connection using its stdin/stdout."

    def __init__(self, cmd):
        import subprocess

        self.subp = subprocess.Popen(
            cmd,
            bufsize=0,
            shell=True,
            preexec_fn=os.setsid,
            stdin=subprocess.PIPE,
            stdout=subprocess.PIPE,
        )

        # Initially was implemented with selectors, but that adds Python3
        # dependency. However, there can be race conditions communicating
        # with a particular child process (like QEMU), and selectors may
        # still work better in that case, so left inplace for now.
        #
        # import selectors
        # self.sel = selectors.DefaultSelector()
        # self.sel.register(self.subp.stdout, selectors.EVENT_READ)

        import select

        self.poll = select.poll()
        self.poll.register(self.subp.stdout.fileno())

    def close(self):
        import signal

        os.killpg(os.getpgid(self.subp.pid), signal.SIGTERM)

    def read(self, size=1):
        data = b""
        while len(data) < size:
            data += self.subp.stdout.read(size - len(data))
        return data

    def write(self, data):
        self.subp.stdin.write(data)
        return len(data)

    def inWaiting(self):
        # res = self.sel.select(0)
        res = self.poll.poll(0)
        if res:
            return 1
        return 0


class ProcessPtyToTerminal:
    """Execute a process which creates a PTY and prints slave PTY as
    first line of its output, and emulate serial connection using
    this PTY."""

    def __init__(self, cmd):
        import subprocess
        import re
        import serial

        self.subp = subprocess.Popen(
            cmd.split(),
            bufsize=0,
            shell=False,
            preexec_fn=os.setsid,
            stdin=subprocess.PIPE,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
        )
        pty_line = self.subp.stderr.readline().decode("utf-8")
        m = re.search(r"/dev/pts/[0-9]+", pty_line)
        if not m:
            print("Error: unable to find PTY device in startup line:", pty_line)
            self.close()
            sys.exit(1)
        pty = m.group()
        # rtscts, dsrdtr params are to workaround pyserial bug:
        # http://stackoverflow.com/questions/34831131/pyserial-does-not-play-well-with-virtual-port
        self.serial = serial.Serial(pty, interCharTimeout=1, rtscts=True, dsrdtr=True)

    def close(self):
        import signal

        os.killpg(os.getpgid(self.subp.pid), signal.SIGTERM)

    def read(self, size=1):
        return self.serial.read(size)

    def write(self, data):
        return self.serial.write(data)

    def inWaiting(self):
        return self.serial.inWaiting()


class Pyboard:
    def __init__(
        self, device, baudrate=115200, user="micro", password="python", wait=0, exclusive=True
    ):
        self.in_raw_repl = False
        self.use_raw_paste = True
        if device.startswith("exec:"):
            self.serial = ProcessToSerial(device[len("exec:") :])
        elif device.startswith("execpty:"):
            self.serial = ProcessPtyToTerminal(device[len("qemupty:") :])
        elif device and device[0].isdigit() and device[-1].isdigit() and device.count(".") == 3:
            # device looks like an IP address
            self.serial = TelnetToSerial(device, user, password, read_timeout=10)
        else:
            import serial
            import serial.tools.list_ports

            # Set options, and exclusive if pyserial supports it
            serial_kwargs = {"baudrate": baudrate, "interCharTimeout": 1}
            if serial.__version__ >= "3.3":
                serial_kwargs["exclusive"] = exclusive

            delayed = False
            for attempt in range(wait + 1):
                try:
                    if os.name == "nt":
                        self.serial = serial.Serial(**serial_kwargs)
                        self.serial.port = device
                        portinfo = list(serial.tools.list_ports.grep(device))  # type: ignore
                        if portinfo and portinfo[0].manufacturer != "Microsoft":
                            # ESP8266/ESP32 boards use RTS/CTS for flashing and boot mode selection.
                            # DTR False: to avoid using the reset button will hang the MCU in bootloader mode
                            # RTS False: to prevent pulses on rts on serial.close() that would POWERON_RESET an ESPxx
                            self.serial.dtr = False  # DTR False = gpio0 High = Normal boot
                            self.serial.rts = False  # RTS False = EN High = MCU enabled
                        self.serial.open()
                    else:
                        self.serial = serial.Serial(device, **serial_kwargs)
                    break
                except (OSError, IOError):  # Py2 and Py3 have different errors
                    if wait == 0:
                        continue
                    if attempt == 0:
                        sys.stdout.write("Waiting {} seconds for pyboard ".format(wait))
                        delayed = True
                time.sleep(1)
                sys.stdout.write(".")
                sys.stdout.flush()
            else:
                if delayed:
                    print("")
                raise PyboardError("failed to access " + device)
            if delayed:
                print("")

    def close(self):
        self.serial.close()

    def read_until(self, min_num_bytes, ending, timeout=10, data_consumer=None):
        # if data_consumer is used then data is not accumulated and the ending must be 1 byte long
        assert data_consumer is None or len(ending) == 1

        data = self.serial.read(min_num_bytes)
        if data_consumer:
            data_consumer(data)
        timeout_count = 0
        while True:
            if data.endswith(ending):
                break
            elif self.serial.inWaiting() > 0:
                new_data = self.serial.read(1)
                if data_consumer:
                    data_consumer(new_data)
                    data = new_data
                else:
                    data = data + new_data
                timeout_count = 0
            else:
                timeout_count += 1
                if timeout is not None and timeout_count >= 100 * timeout:
                    break
                time.sleep(0.01)
        return data

    def enter_raw_repl(self, soft_reset=True):
        self.serial.write(b"\r\x03\x03")  # ctrl-C twice: interrupt any running program

        # flush input (without relying on serial.flushInput())
        n = self.serial.inWaiting()
        while n > 0:
            self.serial.read(n)
            n = self.serial.inWaiting()

        self.serial.write(b"\r\x01")  # ctrl-A: enter raw REPL

        if soft_reset:
            data = self.read_until(1, b"raw REPL; CTRL-B to exit\r\n>")
            if not data.endswith(b"raw REPL; CTRL-B to exit\r\n>"):
                print(data)
                raise PyboardError("could not enter raw repl")

            self.serial.write(b"\x04")  # ctrl-D: soft reset

            # Waiting for "soft reboot" independently to "raw REPL" (done below)
            # allows boot.py to print, which will show up after "soft reboot"
            # and before "raw REPL".
            data = self.read_until(1, b"soft reboot\r\n")
            if not data.endswith(b"soft reboot\r\n"):
                print(data)
                raise PyboardError("could not enter raw repl")

        data = self.read_until(1, b"raw REPL; CTRL-B to exit\r\n")
        if not data.endswith(b"raw REPL; CTRL-B to exit\r\n"):
            print(data)
            raise PyboardError("could not enter raw repl")

        self.in_raw_repl = True

    def exit_raw_repl(self):
        self.serial.write(b"\r\x02")  # ctrl-B: enter friendly REPL
        self.in_raw_repl = False

    def follow(self, timeout, data_consumer=None):
        # wait for normal output
        data = self.read_until(1, b"\x04", timeout=timeout, data_consumer=data_consumer)
        if not data.endswith(b"\x04"):
            raise PyboardError("timeout waiting for first EOF reception")
        data = data[:-1]

        # wait for error output
        data_err = self.read_until(1, b"\x04", timeout=timeout)
        if not data_err.endswith(b"\x04"):
            raise PyboardError("timeout waiting for second EOF reception")
        data_err = data_err[:-1]

        # return normal and error output
        return data, data_err

    def raw_paste_write(self, command_bytes):
        # Read initial header, with window size.
        data = self.serial.read(2)
        window_size = struct.unpack("<H", data)[0]
        window_remain = window_size

        # Write out the command_bytes data.
        i = 0
        while i < len(command_bytes):
            while window_remain == 0 or self.serial.inWaiting():
                data = self.serial.read(1)
                if data == b"\x01":
                    # Device indicated that a new window of data can be sent.
                    window_remain += window_size
                elif data == b"\x04":
                    # Device indicated abrupt end.  Acknowledge it and finish.
                    self.serial.write(b"\x04")
                    return
                else:
                    # Unexpected data from device.
                    raise PyboardError("unexpected read during raw paste: {}".format(data))
            # Send out as much data as possible that fits within the allowed window.
            b = command_bytes[i : min(i + window_remain, len(command_bytes))]
            self.serial.write(b)
            window_remain -= len(b)
            i += len(b)

        # Indicate end of data.
        self.serial.write(b"\x04")

        # Wait for device to acknowledge end of data.
        data = self.read_until(1, b"\x04")
        if not data.endswith(b"\x04"):
            raise PyboardError("could not complete raw paste: {}".format(data))

    def exec_raw_no_follow(self, command):
        if isinstance(command, bytes):
            command_bytes = command
        else:
            command_bytes = bytes(command, encoding="utf8")

        # check we have a prompt
        data = self.read_until(1, b">")
        if not data.endswith(b">"):
            raise PyboardError("could not enter raw repl")

        if self.use_raw_paste:
            # Try to enter raw-paste mode.
            self.serial.write(b"\x05A\x01")
            data = self.serial.read(2)
            if data == b"R\x00":
                # Device understood raw-paste command but doesn't support it.
                pass
            elif data == b"R\x01":
                # Device supports raw-paste mode, write out the command using this mode.
                return self.raw_paste_write(command_bytes)
            else:
                # Device doesn't support raw-paste, fall back to normal raw REPL.
                data = self.read_until(1, b"w REPL; CTRL-B to exit\r\n>")
                if not data.endswith(b"w REPL; CTRL-B to exit\r\n>"):
                    print(data)
                    raise PyboardError("could not enter raw repl")
            # Don't try to use raw-paste mode again for this connection.
            self.use_raw_paste = False

        # Write command using standard raw REPL, 256 bytes every 10ms.
        for i in range(0, len(command_bytes), 256):
            self.serial.write(command_bytes[i : min(i + 256, len(command_bytes))])
            time.sleep(0.01)
        self.serial.write(b"\x04")

        # check if we could exec command
        data = self.serial.read(2)
        if data != b"OK":
            raise PyboardError("could not exec command (response: %r)" % data)

    def exec_raw(self, command, timeout=10, data_consumer=None):
        self.exec_raw_no_follow(command)
        return self.follow(timeout, data_consumer)

    def eval(self, expression, parse=False):
        if parse:
            ret = self.exec_("print(repr({}))".format(expression))
            ret = ret.strip()
            return ast.literal_eval(ret.decode())
        else:
            ret = self.exec_("print({})".format(expression))
            ret = ret.strip()
            return ret

    # In Python3, call as pyboard.exec(), see the setattr call below.
    def exec_(self, command, data_consumer=None):
        ret, ret_err = self.exec_raw(command, data_consumer=data_consumer)
        if ret_err:
            raise PyboardError("exception", ret, ret_err)
        return ret

    def execfile(self, filename):
        with open(filename, "rb") as f:
            pyfile = f.read()
        return self.exec_(pyfile)

    def get_time(self):
        t = str(self.eval("pyb.RTC().datetime()"), encoding="utf8")[1:-1].split(", ")
        return int(t[4]) * 3600 + int(t[5]) * 60 + int(t[6])

    def fs_exists(self, src):
        try:
            self.exec_("import os\nos.stat(%s)" % (("'%s'" % src) if src else ""))
            return True
        except PyboardError:
            return False

    def fs_ls(self, src):
        cmd = (
            "import os\nfor f in os.ilistdir(%s):\n"
            " print('{:12} {}{}'.format(f[3]if len(f)>3 else 0,f[0],'/'if f[1]&0x4000 else ''))"
            % (("'%s'" % src) if src else "")
        )
        self.exec_(cmd, data_consumer=stdout_write_bytes)

    def fs_listdir(self, src=""):
        buf = bytearray()

        def repr_consumer(b):
            buf.extend(b.replace(b"\x04", b""))

        cmd = "import os\nfor f in os.ilistdir(%s):\n" " print(repr(f), end=',')" % (
            ("'%s'" % src) if src else ""
        )
        try:
            buf.extend(b"[")
            self.exec_(cmd, data_consumer=repr_consumer)
            buf.extend(b"]")
        except PyboardError as e:
            raise e.convert(src)

        return [
            listdir_result(*f) if len(f) == 4 else listdir_result(*(f + (0,)))
            for f in ast.literal_eval(buf.decode())
        ]

    def fs_stat(self, src):
        try:
            self.exec_("import os")
            return os.stat_result(self.eval("os.stat(%s)" % ("'%s'" % src), parse=True))
        except PyboardError as e:
            raise e.convert(src)

    def fs_cat(self, src, chunk_size=256):
        cmd = (
            "with open('%s') as f:\n while 1:\n"
            "  b=f.read(%u)\n  if not b:break\n  print(b,end='')" % (src, chunk_size)
        )
        self.exec_(cmd, data_consumer=stdout_write_bytes)

    def fs_readfile(self, src, chunk_size=256):
        buf = bytearray()

        def repr_consumer(b):
            buf.extend(b.replace(b"\x04", b""))

        cmd = (
            "with open('%s', 'rb') as f:\n while 1:\n"
            "  b=f.read(%u)\n  if not b:break\n  print(b,end='')" % (src, chunk_size)
        )
        try:
            self.exec_(cmd, data_consumer=repr_consumer)
        except PyboardError as e:
            raise e.convert(src)
        return ast.literal_eval(buf.decode())

    def fs_writefile(self, dest, data, chunk_size=256):
        self.exec_("f=open('%s','wb')\nw=f.write" % dest)
        while data:
            chunk = data[:chunk_size]
            self.exec_("w(" + repr(chunk) + ")")
            data = data[len(chunk) :]
        self.exec_("f.close()")

    def fs_cp(self, src, dest, chunk_size=256, progress_callback=None):
        if progress_callback:
            src_size = self.fs_stat(src).st_size
            written = 0
        self.exec_("fr=open('%s','rb')\nr=fr.read\nfw=open('%s','wb')\nw=fw.write" % (src, dest))
        while True:
            data_len = int(self.exec_("d=r(%u)\nw(d)\nprint(len(d))" % chunk_size))
            if not data_len:
                break
            if progress_callback:
                written += data_len
                progress_callback(written, src_size)
        self.exec_("fr.close()\nfw.close()")

    def fs_get(self, src, dest, chunk_size=256, progress_callback=None):
        if progress_callback:
            src_size = self.fs_stat(src).st_size
            written = 0
        self.exec_("f=open('%s','rb')\nr=f.read" % src)
        with open(dest, "wb") as f:
            while True:
                data = bytearray()
                self.exec_("print(r(%u))" % chunk_size, data_consumer=lambda d: data.extend(d))
                assert data.endswith(b"\r\n\x04")
                try:
                    data = ast.literal_eval(str(data[:-3], "ascii"))
                    if not isinstance(data, bytes):
                        raise ValueError("Not bytes")
                except (UnicodeError, ValueError) as e:
                    raise PyboardError("fs_get: Could not interpret received data: %s" % str(e))
                if not data:
                    break
                f.write(data)
                if progress_callback:
                    written += len(data)
                    progress_callback(written, src_size)
        self.exec_("f.close()")

    def fs_put(self, src, dest, chunk_size=256, progress_callback=None):
        if progress_callback:
            src_size = os.path.getsize(src)
            written = 0
        self.exec_("f=open('%s','wb')\nw=f.write" % dest)
        with open(src, "rb") as f:
            while True:
                data = f.read(chunk_size)
                if not data:
                    break
                if sys.version_info < (3,):
                    self.exec_("w(b" + repr(data) + ")")
                else:
                    self.exec_("w(" + repr(data) + ")")
                if progress_callback:
                    written += len(data)
                    progress_callback(written, src_size)
        self.exec_("f.close()")

    def fs_mkdir(self, dir):
        self.exec_("import os\nos.mkdir('%s')" % dir)

    def fs_rmdir(self, dir):
        self.exec_("import os\nos.rmdir('%s')" % dir)

    def fs_rm(self, src):
        self.exec_("import os\nos.remove('%s')" % src)

    def fs_touch(self, src):
        self.exec_("f=open('%s','a')\nf.close()" % src)


# in Python2 exec is a keyword so one must use "exec_"
# but for Python3 we want to provide the nicer version "exec"
setattr(Pyboard, "exec", Pyboard.exec_)


def execfile(filename, device="/dev/ttyACM0", baudrate=115200, user="micro", password="python"):
    pyb = Pyboard(device, baudrate, user, password)
    pyb.enter_raw_repl()
    output = pyb.execfile(filename)
    stdout_write_bytes(output)
    pyb.exit_raw_repl()
    pyb.close()


def filesystem_command(pyb, args, progress_callback=None, verbose=False):
    def fname_remote(src):
        if src.startswith(":"):
            src = src[1:]
        # Convert all path separators to "/", because that's what a remote device uses.
        return src.replace(os.path.sep, "/")

    def fname_cp_dest(src, dest):
        _, src = os.path.split(src)
        if dest is None or dest == "":
            dest = src
        elif dest == ".":
            dest = "./" + src
        elif dest.endswith("/"):
            dest += src
        return dest

    cmd = args[0]
    args = args[1:]
    try:
        if cmd == "cp":
            if len(args) == 1:
                raise PyboardError(
                    "cp: missing destination file operand after '{}'".format(args[0])
                )
            srcs = args[:-1]
            dest = args[-1]
            if dest.startswith(":"):
                op_remote_src = pyb.fs_cp
                op_local_src = pyb.fs_put
            else:
                op_remote_src = pyb.fs_get
                op_local_src = lambda src, dest, **_: __import__("shutil").copy(src, dest)
            for src in srcs:
                if verbose:
                    print("cp %s %s" % (src, dest))
                if src.startswith(":"):
                    op = op_remote_src
                else:
                    op = op_local_src
                src2 = fname_remote(src)
                dest2 = fname_cp_dest(src2, fname_remote(dest))
                op(src2, dest2, progress_callback=progress_callback)
        else:
            ops = {
                "cat": pyb.fs_cat,
                "ls": pyb.fs_ls,
                "mkdir": pyb.fs_mkdir,
                "rm": pyb.fs_rm,
                "rmdir": pyb.fs_rmdir,
                "touch": pyb.fs_touch,
            }
            if cmd not in ops:
                raise PyboardError("'{}' is not a filesystem command".format(cmd))
            if cmd == "ls" and not args:
                args = [""]
            for src in args:
                src = fname_remote(src)
                if verbose:
                    print("%s :%s" % (cmd, src))
                ops[cmd](src)
    except PyboardError as er:
        if len(er.args) > 1:
            print(str(er.args[2], "ascii"))
        else:
            print(er)
        pyb.exit_raw_repl()
        pyb.close()
        sys.exit(1)


_injected_import_hook_code = """\
import os, io
class _FS:
  class File(io.IOBase):
    def __init__(self):
      self.off = 0
    def ioctl(self, request, arg):
      return 0
    def readinto(self, buf):
      buf[:] = memoryview(_injected_buf)[self.off:self.off + len(buf)]
      self.off += len(buf)
      return len(buf)
  mount = umount = chdir = lambda *args: None
  def stat(self, path):
    if path == '_injected.mpy':
      return tuple(0 for _ in range(10))
    else:
      raise OSError(-2) # ENOENT
  def open(self, path, mode):
    return self.File()
os.mount(_FS(), '/_')
os.chdir('/_')
from _injected import *
os.umount('/_')
del _injected_buf, _FS
"""


def main():
    import argparse

    cmd_parser = argparse.ArgumentParser(description="Run scripts on the pyboard.")
    cmd_parser.add_argument(
        "-d",
        "--device",
        default=os.environ.get("PYBOARD_DEVICE", "/dev/ttyACM0"),
        help="the serial device or the IP address of the pyboard",
    )
    cmd_parser.add_argument(
        "-b",
        "--baudrate",
        default=os.environ.get("PYBOARD_BAUDRATE", "115200"),
        help="the baud rate of the serial device",
    )
    cmd_parser.add_argument("-u", "--user", default="micro", help="the telnet login username")
    cmd_parser.add_argument("-p", "--password", default="python", help="the telnet login password")
    cmd_parser.add_argument("-c", "--command", help="program passed in as string")
    cmd_parser.add_argument(
        "-w",
        "--wait",
        default=0,
        type=int,
        help="seconds to wait for USB connected board to become available",
    )
    group = cmd_parser.add_mutually_exclusive_group()
    group.add_argument(
        "--soft-reset",
        default=True,
        action="store_true",
        help="Whether to perform a soft reset when connecting to the board [default]",
    )
    group.add_argument(
        "--no-soft-reset",
        action="store_false",
        dest="soft_reset",
    )
    group = cmd_parser.add_mutually_exclusive_group()
    group.add_argument(
        "--follow",
        action="store_true",
        default=None,
        help="follow the output after running the scripts [default if no scripts given]",
    )
    group.add_argument(
        "--no-follow",
        action="store_false",
        dest="follow",
    )
    group = cmd_parser.add_mutually_exclusive_group()
    group.add_argument(
        "--exclusive",
        action="store_true",
        default=True,
        help="Open the serial device for exclusive access [default]",
    )
    group.add_argument(
        "--no-exclusive",
        action="store_false",
        dest="exclusive",
    )
    cmd_parser.add_argument(
        "-f",
        "--filesystem",
        action="store_true",
        help="perform a filesystem action: "
        "cp local :device | cp :device local | cat path | ls [path] | rm path | mkdir path | rmdir path",
    )
    cmd_parser.add_argument("files", nargs="*", help="input files")
    args = cmd_parser.parse_args()

    # open the connection to the pyboard
    try:
        pyb = Pyboard(
            args.device, args.baudrate, args.user, args.password, args.wait, args.exclusive
        )
    except PyboardError as er:
        print(er)
        sys.exit(1)

    # run any command or file(s)
    if args.command is not None or args.filesystem or len(args.files):
        # we must enter raw-REPL mode to execute commands
        # this will do a soft-reset of the board
        try:
            pyb.enter_raw_repl(args.soft_reset)
        except PyboardError as er:
            print(er)
            pyb.close()
            sys.exit(1)

        def execbuffer(buf):
            try:
                if args.follow is None or args.follow:
                    ret, ret_err = pyb.exec_raw(
                        buf, timeout=None, data_consumer=stdout_write_bytes
                    )
                else:
                    pyb.exec_raw_no_follow(buf)
                    ret_err = None
            except PyboardError as er:
                print(er)
                pyb.close()
                sys.exit(1)
            except KeyboardInterrupt:
                sys.exit(1)
            if ret_err:
                pyb.exit_raw_repl()
                pyb.close()
                stdout_write_bytes(ret_err)
                sys.exit(1)

        # do filesystem commands, if given
        if args.filesystem:
            filesystem_command(pyb, args.files, verbose=True)
            del args.files[:]

        # run the command, if given
        if args.command is not None:
            execbuffer(args.command.encode("utf-8"))

        # run any files
        for filename in args.files:
            with open(filename, "rb") as f:
                pyfile = f.read()
                if filename.endswith(".mpy") and pyfile[0] == ord("M"):
                    pyb.exec_("_injected_buf=" + repr(pyfile))
                    pyfile = _injected_import_hook_code
                execbuffer(pyfile)

        # exiting raw-REPL just drops to friendly-REPL mode
        pyb.exit_raw_repl()

    # if asked explicitly, or no files given, then follow the output
    if args.follow or (args.command is None and not args.filesystem and len(args.files) == 0):
        try:
            ret, ret_err = pyb.follow(timeout=None, data_consumer=stdout_write_bytes)
        except PyboardError as er:
            print(er)
            sys.exit(1)
        except KeyboardInterrupt:
            sys.exit(1)
        if ret_err:
            pyb.close()
            stdout_write_bytes(ret_err)
            sys.exit(1)

    # close the connection to the pyboard
    pyb.close()


if __name__ == "__main__":
    main()


# Copyright (C) 2024 Dr. Matthias Kesenheimer - All Rights Reserved.
# You may use, distribute and modify this code under the terms of the GPL3 license.
#
# You should have received a copy of the GPL3 license with this file.
# If not, please write to: info@faultyhardware.de.

from Spline import Spline

class PulseGenerator():
    @micropython.native
    def __init__(self, time_resolution:int = 10, vhigh:float = 1.0, factor:float = 1.0):
        """
        TODO
        """
        # all units in [ns] and [V]
        if time_resolution < 6:
            raise Exception("Error: time_resolution too small.")
        self.time_resolution = time_resolution
        self.frequency = 1_000_000_000 / self.time_resolution
        self.max_points = 4096
        self.points_per_ns = 1 / self.time_resolution
        self.set_calibration(output_voltage_at_minimal_gain=vhigh, calibration_factor=factor)
        self.set_offset(offset=3.3)
        # caching
        self.total_pulse_duration = 0
        self.pulse = []
        # coefficients for spline interpolation
        self.coefficients = None
        self.grid_hat = None

    @micropython.native
    def get_points_per_ns(self):
        return self.points_per_ns

    @micropython.native
    def calculate_pulse_number_of_points(self, total_pulse_duration:int) -> int:
        """
        TODO
        """
        pulse_number_of_points = int(total_pulse_duration / self.time_resolution)
        if pulse_number_of_points > self.max_points:
            return self.max_points
        return pulse_number_of_points

    @micropython.native
    def calibration_pulse(self) -> list[int]:
        high = int((0) * self.points_per_volt)
        low = int((0 - self.offset) * self.points_per_volt)
        pulse = [high] * 1000 + [low] * 1000
        return pulse

    @micropython.native
    def set_offset(self, offset:float):
        """
        TODO
        """
        self.offset = offset
        self.gain = self.offset / self.output_voltage_at_minimal_gain
        voltage_resolution = 4096
        self.points_per_volt = int(self.calibration_factor * voltage_resolution / self.gain)

    @micropython.native
    def get_value(self, voltage:float):
        v = voltage - self.offset
        return int(v * self.points_per_volt)

    @micropython.native
    def set_calibration(self, output_voltage_at_minimal_gain:float, calibration_factor:float):
        self.output_voltage_at_minimal_gain = output_voltage_at_minimal_gain
        self.calibration_factor = calibration_factor

    @micropython.native
    def get_frequency(self):
        """
        TODO
        """
        return self.frequency

    @micropython.native
    def get_max_points(self):
        """
        TODO
        """
        return self.max_points

    @micropython.native
    def pulse_from_config(self, ps_config:list[list[float]], padding:bool = False) -> list[int]:
        """
        TODO
        """
        pulse = []
        for point in ps_config:
            t = point[0]
            v = point[1] - self.offset
            n = int(t * self.points_per_ns)
            value = int(v * self.points_per_volt)
            pulse += [value] * n
        # padding with last value
        if padding:
            length = len(pulse)
            if length < self.max_points:
                last_value = pulse[-1]
                pulse += [last_value] * (self.max_points - length)
        # sanity check
        if len(pulse) > self.max_points:
            raise Exception("Erroneous pulse config: pulse too large.")
        return pulse

    @micropython.native
    def pulse_from_spline(self, xpoints:list[int], ypoints:list[float], padding:bool = False) -> list[int]:
        """
        TODO
        """
        if len(xpoints) != len(ypoints):
            raise Exception("xpoints and ypoints have different lengths.")
        tpoints = [0] * len(xpoints)
        vpoints = [0] * len(xpoints)
        for i in range(len(xpoints)):
            tpoints[i] = int(xpoints[i] * self.points_per_ns)
            vpoints[i] = int((ypoints[i] - self.offset) * self.points_per_volt)
        a = tpoints[0]
        b = tpoints[-1]
        if a == b:
            return vpoints
        #print(f"frequency = {self.frequency}")
        #print(f"tpoints = {tpoints}")
        #print(f"vpoints = {vpoints}")
        #print(f"a = {a}")
        #print(f"b = {b}")
        #self.coefficients = Spline.cal_coefs(a, b, vpoints)
        self.grid_hat = Spline.calc_grid(a, b, b - a) # grid with step size one
        #self.pulse = list([int(Spline.interpolate(x, a, b, self.coefficients)) for x in self.grid_hat])
        self.pulse = Spline.pchip_interpolate(tpoints, vpoints, self.grid_hat)
        self.pulse = list(map(int, self.pulse))
        #print(f"offset = {self.offset}")
        #print(f"points_per_volt = {self.points_per_volt}")
        #print(f"points_per_ns = {self.points_per_ns}")
        return self.pulse

    @micropython.native
    def pulse_from_lambda(self, ps_lambda, total_pulse_duration:int, padding:bool = False) -> list[int]:
        """
        TODO
        """
        if self.total_pulse_duration != total_pulse_duration:
            pulse_number_of_points = self.calculate_pulse_number_of_points(total_pulse_duration)
            self.pulse = [0] * pulse_number_of_points
        t = 0
        dt = self.time_resolution
        for i in range(pulse_number_of_points):
            self.pulse[i] = int((ps_lambda(t) - self.offset) * self.points_per_volt)
            t += dt
        # padding with last value
        if padding:
            if pulse_number_of_points < self.max_points:
                last_value = self.pulse[-1]
                self.pulse += [last_value] * (self.max_points - pulse_number_of_points)
        return self.pulse

    @micropython.native
    def pulse_from_list(self, pulse:list[int], padding:bool = False) -> list[int]:
        """
        Puls is generated from raw list without offset or gain correction applied.
        """
        # padding with last value
        if padding:
            length = len(pulse)
            if length < self.max_points:
                last_value = pulse[-1]
                pulse += [last_value] * (self.max_points - length)
        # sanity check
        if len(pulse) > self.max_points:
            raise Exception("Fatal error: pulse too large.")
        return pulse


#!/usr/bin/env python3
# Copyright (C) 2025 Dr. Matthias Kesenheimer - All Rights Reserved.
# You may use, distribute and modify this code under the terms of the GPL3 license.
#
# You should have received a copy of the GPL3 license with this file.
# If not, please write to: info@faultyhardware.de.

# SQL Queries:
# Show only successes and flash-resets:
# color = 'R' or response LIKE '_Warning.flash_reset'

import argparse
import sys
import os.path

# import custom libraries
from findus import Database


class DBTool:
    def __init__(self, args):
        self.args = args

    def run(self):
        if self.args.dbname is not None and self.args.cleanup is not None:
            if not os.path.isfile(self.args.dbname):
                print(f"[-] Database {self.args.dbname} does not exist.")
                sys.exit(1)
            database = Database(sys.argv, dbname=self.args.dbname, dirname=None)
            database.cleanup(self.args.cleanup)

        if self.args.dbname is not None and self.args.remove is not None:
            if not os.path.isfile(self.args.dbname):
                print(f"[-] Database {self.args.dbname} does not exist.")
                sys.exit(1)
            database = Database(sys.argv, dbname=self.args.dbname, dirname=None, resume=True)
            database.remove_conditional(self.args.remove)

        if self.args.merge is not None:
            database1 = Database(sys.argv, dbname=self.args.merge[0], dirname=None, resume=True)
            database2 = Database(sys.argv, dbname=self.args.merge[1], dirname=None, resume=True)
            if os.path.isfile(self.args.merge[2]):
                print(f"[-] Database {self.args.merge[2]} already exists. Choose another name.")
                sys.exit(1)
            column_names1 = database1.get_column_names()
            column_names2 = database2.get_column_names()
            if column_names1 != column_names2:
                print("[-] Column names of databases not compatible.")
                sys.exit(1)
            out = Database(sys.argv, dbname=self.args.merge[2], dirname=None, column_names=column_names1)
            toteid = 0
            for eid in range(0, database1.get_base_experiments_count()):
                parameters = database1.get_parameters_of_experiment(eid)
                if parameters != [None]:
                    parameters = (toteid, ) + parameters[1:]
                    print(parameters)
                    out.insert(*parameters)
                    toteid += 1
            for eid in range(0, database2.get_base_experiments_count()):
                parameters = database2.get_parameters_of_experiment(eid)
                if parameters != [None]:
                    parameters = (toteid, ) + parameters[1:]
                    print(parameters)
                    out.insert(*parameters)
                    toteid += 1

def main(argv=sys.argv):
    parser = argparse.ArgumentParser()
    parser.add_argument("--dbname", required=False, default=None, help="Name of the database")
    parser.add_argument("--cleanup", required=False, default=None, help="Remove experiments with a certain color from the database.")
    parser.add_argument("--remove", required=False, default=None, help="Remove experiments by a certain condition. For example, `id = 1000`, `id > 1000`, `color = \"C\"` or `delay < 6000`.")
    parser.add_argument("--merge", metavar=("databas1", "database2", "output"), required=False, default=None, nargs=3, help="Merge two databases.")

    args = parser.parse_args()

    dbtool = DBTool(args)

    try:
        dbtool.run()
    except KeyboardInterrupt:
        print("\nExitting...")
        sys.exit(0)

if __name__ == "__main__":
    main()



#!/usr/bin/env python3
# Copyright (C) 2024 Dr. Matthias Kesenheimer - All Rights Reserved.
# You may use, distribute and modify this code under the terms of the GPL3 license.
#
# You should have received a copy of the GPL3 license with this file.
# If not, please write to: info@faultyhardware.de.

"""
GeneticAlgorithm

This module includes all classes and functions that are necessary to search for optimal parameters in a multidimenstional parameter space.
"""

import random

class Parameterspace():
    def __init__(self, parameter_boundaries:list[tuple[float, float]], parameter_divisions:list[int]):
        """
        Initializes the class with the given parameter boundaries and divisions.
    
        Parameters:
            parameter_boundaries: A list of tuples representing the lower and upper bounds for each parameter.
            parameter_divisions: A list of integers representing the number of divisions for each parameter.
    
        Raises:
            Exception: If the lengths of parameter_boundaries and parameter_divisions are different.
            Exception: If any upper bound is lower than the corresponding lower bound.
        """
        # sanity checks
        if len(parameter_boundaries) != len(parameter_divisions):
            raise Exception(f"Error: parameter_boundaries and parameter_divisions have different lengths ({len(parameter_boundaries)} vs. {len(parameter_divisions)}).")
        for tup in parameter_boundaries:
            if tup[0] > tup[1]:
                raise Exception(f"Error: Upper bound {tup[1]} is lower than lower bound {tup[0]}.")
        self.parameter_boundaries = parameter_boundaries
        self.parameter_divisions = parameter_divisions
        self.cardinality = 1
        for num in self.parameter_divisions:
            self.cardinality *= num
        self.weights_per_bin = [0 for x in range(self.cardinality)]

    def get_cardinality(self):
        """
        Returns the cardinality of the set.
        """
        return self.cardinality

    def get_bin_assignment(self, *parameter:float) -> list[int]:
        """
        This function assigns a bin to each parameter based on its value.
        
        Parameters:
            *parameter: A list of float values representing the parameters to be assigned to bins.
        
        Returns:
            list: A list of integers representing the bin numbers for each parameter.
        """
        fact = 1
        bina = 0
        for i in range(len(self.parameter_divisions)):
            division =  self.parameter_divisions[i]
            # xdelta = (xmax - xmin) / xdiv
            delta = (self.parameter_boundaries[i][1] - self.parameter_boundaries[i][0]) / division
            bin_number = int((parameter[i] - self.parameter_boundaries[i][0]) / delta)
            #print(f"par = {parameter[i]}, delta = {delta}, bina = {bin_number}")
            if i > 0:
                fact *= self.parameter_divisions[i - 1]
            bina += fact * bin_number
        return bina

    def add_experiment(self, weight:int, *parameter:float):
        """
        Add experiment data to the dataset.
        This function adds an experiment to the dataset with the given weight and parameters.
        It checks if each parameter is within the specified boundaries.
        If any parameter is out of bounds, it prints an error message and returns without
        adding the experiment. Otherwise, it assigns the experiment to a bin and updates the weight for that bin.
    
        Parameters:
            weight: The weight to assign to the experiment.
            *parameter: The parameters for the experiment, must be within the specified boundaries.
    
        Returns:
            None
        """
        for i in range(len(self.parameter_divisions)):
            if parameter[i] < self.parameter_boundaries[i][0] or parameter[i] >= self.parameter_boundaries[i][1]:
                print("[-] Error: parameter out of bounds. Skipping.")
                print(f"[-] Dimension: {i}, lower = {self.parameter_boundaries[i][0]}, upper = {self.parameter_boundaries[i][1]}, parameter = {parameter[i]}")
                return
        bina = self.get_bin_assignment(*parameter)
        self.weights_per_bin[bina] += weight

    def get_weights(self) -> list[int]:
        """
        Returns the list of weights per bin.
    
        Returns:
            list: A list of weights corresponding to each bin.
        """
        return self.weights_per_bin

    def get_bin_numbers_sorted_by_weights(self) -> list[int]:
        """
        Returns a list of bin numbers sorted by their corresponding weights.
    
        Returns:
            list: A list of bin numbers sorted by increasing weights.
        """
        return sorted(range(len(self.weights_per_bin)), key=lambda i: self.weights_per_bin[i])

    def get_coordinates(self, bin_assignment:int) -> list[int]:
        """
        This function calculates the coordinates of a given bin assignment in a multi-dimensional binning system.
        
        Parameters:
            bin_assignment: The bin assignment for which to calculate the coordinates.
        
        Returns:
            list: A list of coordinates corresponding to the bin assignment.
        
        Raises:
            Exception: If the bin assignment exceeds the total number of bins.
        """
        bin_numbers = []
        reversed_parameter_divisions = list(reversed(self.parameter_divisions))
        fact = self.cardinality
        if bin_assignment > fact:
            raise Exception("Error: bin number exceeds total number of bins.")
        for div in reversed_parameter_divisions:
            c = (fact / div)
            b = int(bin_assignment / c)
            #print(f"bin_assignment = {bin_assignment}, fact = {fact}, div = {div}, (fact / div) = {c}, bin_number = {b}")
            bin_numbers.append(b)
            fact /= div
            bin_assignment -= int(b * c)
        return list(reversed(bin_numbers))

    def get_boundaries_from_coordinates(self, coordinates:list[float]) -> list[tuple[float, float]]:
        """
        This function calculates the boundaries for each parameter based on the given coordinates.
        
        Parameters:
            coordinates: A list of coordinates.
        
        Returns:
            list: A list of tuples representing the boundaries for each parameter.
        """
        boundaries = []
        for i in range(len(self.parameter_divisions)):
            division =  self.parameter_divisions[i]
            delta = (self.parameter_boundaries[i][1] - self.parameter_boundaries[i][0]) / division
            lower = self.parameter_boundaries[i][0] + delta * coordinates[i]
            upper = self.parameter_boundaries[i][0] + delta * (coordinates[i] + 1)
            tup = (lower, upper)
            boundaries.append(tup)
        return boundaries

    def get_boundaries(self, bin_assignment:int) -> list[tuple[float, float]]:
        """
        Returns the boundaries for a given bin assignment.

        Parameters:
            bin_assignment: The bin assignment for which to get the boundaries.

        Returns:
            list: A list of tuples representing the boundaries.
        """
        bin_numbers = self.get_coordinates(bin_assignment)
        return self.get_boundaries_from_coordinates(bin_numbers)

class Individual():
    """
    This class represents an individual with parameters, health, maximum age, and current age.

    Attributes:
        parameters: A list of parameters associated with the individual.
        health: The current health status of the individual.
        max_age: The maximum age that the individual can reach.
        age: The current age of the individual.
    """
    def __init__(self, parameters:list[int], max_age:int = 10):
        self.parameters = parameters
        self.health = 0
        self.max_age = max_age
        self.age = 0

    def set_genom(self, parameters: list[int]):
        """
        Set the parameters (genom) for the entity.

        Parameters:
            parameters: The parameters to set.
        """
        self.parameters = parameters
    
    def get_genom(self) -> list[int]:
        """
        Get the parameters (genom) of the entity.
        
        Returns:
            list: The parameters (genom) of the entity.
        """
        return self.parameters
    
    def set_health(self, health: int):
        """
        Set the health of the entity.
        
        Parameters:
            health: The health value to set.
        """
        self.health = health
    
    def get_health(self) -> int:
        """
        Get the health of the entity.
        
        Returns:
            int: The health of the entity.
        """
        return self.health
    
    def get_age(self) -> int:
        """
        Get the current age of the entity.
        
        Returns:
            int: The current age of the entity.
        """
        return self.age
    
    def get_max_age(self) -> int:
        """
        Get the maximum age of the entity.
        
        Returns:
            int: The maximum age of the entity.
        """
        return self.max_age
    
    def increase_age(self):
        """
        Increase the age of the entity by one.
        """
        self.age += 1

class Population():
    """
    A class to represent a population of individuals.
    """
    def __init__(self, number_of_individuals:int, length_of_genom:int):
        """
        Initializes a new population with a given number of individuals and genom length.

        Parameters:
            number_of_individuals: The number of individuals in the population.
            length_of_genom: The length of the genom for each individual.
        """
        if number_of_individuals < 10:
            raise Exception("Error: Population too small.")

        self.number_of_individuals = number_of_individuals
        self.length_of_genom = length_of_genom

        self.population = [None] * self.number_of_individuals
        for i in range(self.number_of_individuals):
            self.population[i] = self.generate_random_individual()

    def get_number_of_individuals(self) -> int:
        """
        Returns the number of individuals in the population.

        Returns:
            The number of individuals.
        """
        return self.number_of_individuals

    def get_length_of_genom(self) -> int:
        """
        Returns the length of the genom for each individual.

        Returns:
            The length of the genom.
        """
        return self.length_of_genom

    def generate_random_individual(self) -> Individual:
        """
        Generates a random individual with a genom of length 'length_of_genom'.

        Returns:
            A random individual.
        """
        genom = [0] * self.length_of_genom
        for i in range(self.length_of_genom):
            genom[i] = random.uniform(0, 1)
        return Individual(genom)

    def get_individuals(self) -> list[Individual]:
        """
        Returns a list of all individuals in the population.

        Returns:
            A list of individuals.
        """
        return self.population

    def set_individuals(self, individuals:list[Individual]):
        """
        Sets the population to the given list of individuals.
        
        Parameters:
            individuals: A list of individuals to set as the population.
        """
        self.population = individuals

    def sort_by_health(self):
        """
        Sorts the population by the health of each individual.
        """
        self.population = sorted(self.population, key=lambda ind: ind.get_health())

    def update_health(self, health_function):
        """
        Updates the health of each individual in the population using the given health function.

        Parameters:
            health_function: A function that takes a genom and returns a health value.
        """
        for ind in self.population:
            ind.set_health(health_function(ind.get_genom()))
            #print(ind.get_health())

    def breed(self, i:int, j:int) -> Individual:
        """
        Breeds two individuals and returns a new individual with a genom that is a combination of the two parents.

        Parameters:
            i: The index of the first parent individual.
            j: The index of the second parent individual.

        Returns:
            A new individual that is a result of breeding the two parents.
        """
        geni = self.population[i].get_genom().copy()
        genj = self.population[j].get_genom().copy()
        for k in range(len(geni)):
            select = random.randint(0, 1)
            if select == 1:
                geni[k] = genj[k]
        return Individual(geni)

    def mutate(self, i:int) -> Individual:
        """
        Mutates the genom of the individual at the given index.

        Parameters:
            i: The index of the individual to mutate.
        
        Returns:
            The mutated individual.
        """
        gene = random.randint(0, self.length_of_genom - 1)
        self.population[i].get_genom()[gene] = random.uniform(0, 1)

    def replace(self, i:int, individual:Individual):
        """
        Replaces the individual at the given index with the given individual.

        Parameters:
            i: The index of the individual to replace.
            individual: The individual to replace the old individual with.
        """
        self.population[i] = individual

    def replace_with_random(self, i:int):
        """
        Replaces the individual at the given index with a random individual.

        Parameters:
            i: The index of the individual to replace.
        """
        self.population[i] = self.generate_random_individual()

    def kill_and_replace(self):
        """
        Removes individuals from the population that have exceeded their maximum age and replaces them with random individuals.
        """
        for i in range(len(self.population)):
            if self.population[i].get_age() > self.population[i].get_max_age():
                self.population[i] = self.generate_random_individual()

    def increase_age_of_population(self):
        """
        Increases the age of each individual in the population.
        """
        for i in range(len(self.population)):
            self.population[i].increase_age()

class GeneticAlgorithm:
    """
    Class representing a Genetic Algorithm for optimizing a given problem.
    """
    def __init__(self, parameterspace:Parameterspace, population:Population, health_malus_factor:float = 1):
        """
        Initializes the Genetic Algorithm with a parameterspace, population, and an optional health malus factor.
        
        Parameters:
            parameterspace: The space of parameters over which the algorithm operates.
            population: The initial population of candidate solutions.
            health_malus_factor: A factor that penalizes solutions based on their healthiness.
        """
        self.parameterspace = parameterspace
        self.population = population
        self.health_malus_factor = health_malus_factor

    def get_bins_from_genom(self, parameters:list[int]) -> list[int]:
        """
        Converts a list of parameters into a list of bins based on the parameterspace.
        
        Parameters:
            parameters: A list of parameters.

        Returns:
            A list of bins corresponding to the parameters.
        """
        bins = [int(x * self.parameterspace.get_cardinality()) for x in parameters]
        return bins

    def health_function(self, parameters:list[int]) -> int:
        """
        Calculates the health of an individual:

            health = Sum(weight_i) - factor * malus * Sum(weight_i)
            health = (1 - factor * malus) * Sum(weight_i)

        Parameters:
            parameters: The parameters to set.
        """
        bins = self.get_bins_from_genom(parameters)
        # sum up the weights in each bin
        health = 0
        weights = self.parameterspace.get_weights()
        for b in bins:
            health += weights[b]

        # for every bin that occurs more than once, reduce health
        # (forces the algorithm to look into separate bins)
        # malus can maximal be the number of genoms (=bins),
        # therefore it is reasonable to choose health_malus_factor < (1 / number_of_bins)
        counts = {}
        for item in bins:
            counts[item] = counts.get(item, 0) + 1
        malus = 0
        for c in counts:
            if counts[c] > 1:
                malus += counts[c] - 1
        health -= (self.health_malus_factor * malus * health)

        return health

    def get_max_health(self) -> int:
        """
        Calculate the maximum health possible for an individual based on the population's genom length and the weights in the parameter space.
    
        Returns:
            Maximum health value.
        """
        genom_length = self.population.get_length_of_genom()
        return max(self.parameterspace.get_weights()) * genom_length

    def step(self) -> list[Individual]:
        """
        Perform one step of the simulation by updating the population, performing breeding, mutation, replacing weak individuals, and killing old individuals.
    
        Returns:
            List of updated individuals in the population.
        """
        self.population.increase_age_of_population()
        self.population.update_health(self.health_function)
        self.population.sort_by_health()

        # Step 1: breeding
        number_of_individuals = self.population.get_number_of_individuals()
        child12 = self.population.breed(number_of_individuals - 1, number_of_individuals - 2)
        child21 = self.population.breed(number_of_individuals - 2, number_of_individuals - 1)
        child34 = self.population.breed(number_of_individuals - 3, number_of_individuals - 4)
        child43 = self.population.breed(number_of_individuals - 4, number_of_individuals - 3)
        # replace the individuals with bad health
        self.population.replace(0, child12)
        self.population.replace(1, child21)
        self.population.replace(2, child34)
        self.population.replace(3, child43)

        # Step 2: mutate
        self.population.mutate(0)
        self.population.mutate(2)

        # Step 3: replace weak individuals with random ones
        self.population.replace_with_random(4)
        self.population.replace_with_random(5)

        # Step 4: Kill old individuals
        self.population.kill_and_replace()

        self.population.update_health(self.health_function)
        self.population.sort_by_health()

        return self.population.get_individuals()

    def run(self, threshold:float) -> list[int]:
        """
        Runs the simulation until an individual's health meets or exceeds a certain threshold.
            
        Parameters:
            threshold: The threshold as a fraction of the maximum health value.
            
        Returns:
            A list of bins derived from the genom of the individual with the highest health.
        """
        maxhealth = self.get_max_health()
        while True:
            individuals = self.step()
            for ind in individuals:
                print(f"health of ind {ind}: {ind.get_health()}")
            print()
            if individuals[0].get_health() >= maxhealth * threshold:
                parameters = individuals[0].get_genom()
                return self.get_bins_from_genom(parameters)

    def get_population(self) -> Population:
        """
        Returns the population.
        
        Returns:
            Population
        """
        return self.population

    def get_parameterspace(self) -> Parameterspace:
        """
        Returns the parameterspace.
        
        Returns:
            Parameterspace
        """
        return self.parameterspace

class OptimizationController():
    """
    Wrapper class for initializing parameter space binning and using the genetic algorithm to search for optimal paramters.

    Methods:
        __init__: Constructor of the OptimizationController. Parameter boundaries and parameter divisions must be provided. Parameters for the genetic algorithm are optional.
        print_best_performing_bins: Output the best performing parameter space bins. In these bins a successful glitch is assumed.
        step: Perform the next step of the genetic algorithm; should be called before every experiment.
        add_experiment: Add the parameters and the outcome (weights) to the parameter space.
    """

    def __init__(self, parameter_boundaries:list[tuple[int, int]], parameter_divisions:list[int], number_of_individuals:int = 10, length_of_genom:int = 20, malus_factor_for_equal_bins:float = 1):
        """
        Constructor of the OptimizationController. Parameter boundaries and parameter divisions must be provided. Parameters for the genetic algorithm are optional.

        Parameters:
            parameter_boundaries: Boundaries of all parameters to look into, for example `[(s_delay, e_delay), (s_t1, e_t1), (s_length, e_length)]`.
            parameter_divisions: Number of how many bins the individual parameter ranges should be divided into, for example `[10, 10, 5]`. Bigger values mean division into smaller bins which increases accuracy at the expense of performance (longer glitching campaign duration).
            number_of_individuals: Number of individuals used simultaneously in the genetic algorithm. Increasing this number improves accuracy at the expense of performance (longer glitching campaign duration).
            length_of_genom: Number of bins to track per individual. The larger the parameter space, the higher this number should be.
            malus_factor_for_equal_bins: Can be used to punish individuals that look into the same bin multiple times. Must be in range [0, 1]. A bigger value means a greater penalty.
        """
        self.par = Parameterspace(parameter_boundaries, parameter_divisions)
        self.pop = Population(number_of_individuals, length_of_genom)
        # renorm the malus factor to the maximum number of genoms,
        # thus malus_factor_for_equal_bins can be chosen between 0 and 1.
        if malus_factor_for_equal_bins > 1 or malus_factor_for_equal_bins < 0:
            print("[-] Error: malus_factor_for_equal_bins must be in range [0, 1].")
            sys.exit(-1)
        factor = malus_factor_for_equal_bins / length_of_genom
        self.opt = GeneticAlgorithm(self.par, self.pop, factor)
        self.i_current_individual = 0
        self.i_current_bin = 0
        self.number_of_individuals = self.pop.get_number_of_individuals()
        self.length_of_genom = self.pop.get_length_of_genom()

    def print_best_performing_bins(self):
        """
        Method that prints an overview of the best performing bins of the parameterspace. Can be called from time to time to obtain an overview of the optimization process.
        """
        number_of_individuals = self.pop.get_number_of_individuals()
        ind0 = self.pop.get_individuals()[number_of_individuals - 1]
        print(f"[+] Individual health = {ind0.get_health()}, age = {ind0.get_age()}")
        print("[+] Best performing bins:")
        genom = ind0.get_genom()
        bins = self.opt.get_bins_from_genom(genom)
        boundaries = []
        for i in range(len(bins)):
            boundary = self.par.get_boundaries(bins[i])
            print(f"    bin = {bins[i]}: {boundary}")
            boundaries.append(boundary)

    def step(self) -> list[int]:
        """
        Perform the next step in the optimization process. Should be called before the next experiment.

        Returns:
            A list of the parameters for the next experiment.
        """
        individuals = self.pop.get_individuals()
        parameters = individuals[self.i_current_individual].get_genom()
        bins = self.opt.get_bins_from_genom(parameters)
        boundaries = self.par.get_boundaries(bins[self.i_current_bin])

        random_numbers = []
        for b in boundaries:
            random_numbers.append(random.uniform(b[0], b[1]))

        # next bin
        self.i_current_bin += 1
        if self.i_current_bin >= self.length_of_genom:
            self.i_current_bin = 0
            # next individual
            self.i_current_individual += 1
            if self.i_current_individual >= self.number_of_individuals:
                self.i_current_individual = 0
                # next age step
                self.pop.set_individuals(self.opt.step())

        return [round(x, 1) for x in random_numbers]

    def add_experiment(self, weight:int, *parameter:int):
        """
        Method to add the parameters and the outcome (weights) to the parameter space. Can be used similarly to the method `Database.insert()`.

        Parameters:
            weight: Weight of the outcome of the experiment. Higher values mean a better performing bin and a better health of the individuals that track that bin.
            parameter: list of parameters that were used for the experiment.
        """
        self.par.add_experiment(weight, *parameter)



#!/usr/bin/env python3
# Copyright (C) 2024 Dr. Matthias Kesenheimer - All Rights Reserved.
# You may use, distribute and modify this code under the terms of the GPL3 license.
#
# You should have received a copy of the GPL3 license with this file.
# If not, please write to: info@faultyhardware.de.

import argparse
import os
import sys
import findus
from findus.helper import upload

def files_for_version(version: str) -> list[str]:
    """
    """
    files = ["AD910X.py", "FastADC.py", "Globals.py", "PicoGlitcher.py", "PulseGenerator.py", "Spline.py", "Statemachines.py"]
    if version == 'v1':
        files.append("config_v1/config.json")
    elif version == 'v2.1' or version == 'v2.2':
        files.append("config_v2.1-2/config.json")
    elif version == 'v2.3' or version == 'v2.4':
        files.append("config_v2.3-4/config.json")
    elif version == 'v2.5' or version == 'v3.0':
        files.append("config_v3.0/config.json")
    else:
        print(f"[-] Version string {version} not allowed. Choose either v1, v2.1, v2.2, v2.3, v2.4, v2.5 or v3.0.")
        sys.exit(-1)
    module_path = os.path.dirname(os.path.abspath(findus.__file__))
    firmware_path = os.path.join(module_path, "firmware")
    print(f"[+] Using base path {firmware_path}")
    files_with_path = [os.path.join(firmware_path, f) for f in files]
    return files_with_path

def main(argv=sys.argv):
    parser = argparse.ArgumentParser(
        description="Update the firmware of the Pico Glitcher."
    )
    parser.add_argument("--port", help="/dev/tty* of the Raspberry Pi Pico", required=True, default='/dev/ttyACM1')
    parser.add_argument("--version", help="Pico Glitcher (one of v1, v2.1, v2.2, v2.3, v2.4, v2.5, v3.0)", required=False, default='v3.0')
    args = parser.parse_args()

    # get the file list
    files = files_for_version(args.version)

    # reset before tasks to have a well defined state
    upload.reset(args.port)

    pyb = upload.connect(args.port)
    remote_files = upload.list_remote_files(pyb)

    print("[+] Deleting all files...")
    upload.delete_files(pyb, remote_files)

    pyb.exit_raw_repl()
    pyb.close()

    upload.upload_files_batch(files, args.port)
    upload.reset(args.port)

    # what is on the Raspberry Pi Pico?
    upload.print_remote_content(args.port)

if __name__ == "__main__":
    main()


#!/usr/bin/env python3
# Copyright (C) 2024 Dr. Matthias Kesenheimer - All Rights Reserved.
# You may use, distribute and modify this code under the terms of the GPL3 license.
#
# You should have received a copy of the GPL3 license with this file.
# If not, please write to: info@faultyhardware.de.

import argparse
import sys
import time
from findus import PicoGlitcher

class PowerCycler:
    def __init__(self, args):
        self.args = args
        if self.args.rpico == "":
            print("[+] Initializing ProGlitcher")
            from findus.ProGlitcher import ProGlitcher
            self.glitcher = ProGlitcher()
            self.glitcher.init()
        else:
            print("[+] Initializing PicoGlitcher")
            self.glitcher = PicoGlitcher()
            self.glitcher.init(port=args.rpico, ext_power=args.power, ext_power_voltage=args.voltage)

        # choose multiplexing, pulse-shaping or crowbar glitching
        if args.multiplexing:
            self.glitcher.change_config_and_reset("mux_vinit", str(args.voltage))
            self.glitcher.init(port=args.rpico, ext_power=args.power, ext_power_voltage=args.voltage)
            self.glitcher.set_multiplexing()
        elif args.pulse_shaping:
            self.glitcher.set_pulseshaping(vinit=args.voltage)

    def run(self):
        print("[+] Power cycling target")
        # power cycle target
        self.glitcher.power_cycle_target(1)
        self.glitcher.reset(0.01)
        time.sleep(0.2)

def main(argv=sys.argv):
    parser = argparse.ArgumentParser(description="Power-cycle the target via different output stages of the Pico Glitcher (VTARGET, multiplexing stage, pulse-shaping stage or external power supply).")
    parser.add_argument("--rpico", required=False, help="rpico port", default="/dev/ttyACM0")
    parser.add_argument("--power", required=False, help="rk6006 port", default=None)
    parser.add_argument("--multiplexing", required=False, action='store_true', help="Use the multiplexing stage to power-cycle the target (requires PicoGlitcher v2).")
    parser.add_argument("--pulse-shaping", required=False, action='store_true', help="Use the pulse-shaping stage to power-cycle the target (requires PicoGlitcher v2). Be sure to calibrate the pulse-shaping stage's voltage output.")
    parser.add_argument("--voltage", required=False, help="The voltage to set. Note that the voltage output of the pulse-shaping stage can not be controlled with this parameter. The voltage output of the pulse-shaping stage must be set manually with the potentiometer.", type=float, default=3.3)
    args = parser.parse_args()

    pc = PowerCycler(args)

    try:
        pc.run()
    except KeyboardInterrupt:
        print("\nExitting...")
        sys.exit(1)

if __name__ == "__main__":
    main()


#!/usr/bin/env python3
# Copyright (C) 2024 Dr. Matthias Kesenheimer - All Rights Reserved.
# You may use, distribute and modify this code under the terms of the GPL3 license.
#
# You should have received a copy of the GPL3 license with this file.
# If not, please write to: info@faultyhardware.de.

import serial
import os
import math
import argparse
from time import sleep
import struct
import sys

def hex_int(value:str):
    return int(value, 16)

class STM8Programmer():
    BLOCK_SIZE = 64 # 128 for parts with >8k flash

    REQ_ENTER = (0xde, 0xad, 0xbe, 0xef, )
    BOOTLOADER_INIT = (0x7f, )
    WRITE_MEMORY = (0x10, )
    READ_MEMORY = (0x11, )
    OPTION_WRITE = (0x12, )
    OPTION_READ = (0x13, )

    ACK  = (0x79, )
    NACK = (0x1f, )

    def __init__(self, port, baud=115200):
        try:
            self.ser = serial.Serial(port, baud, timeout=0.05)
        except:
            print(f"[+] Seriell port {port} not found. Exiting.")
            sys.exit(1)

    def __del__(self):
        try:
            self.ser.setRTS(0)
            self.ser.close()
        except:
            pass

    def crc8_update(self, data, crc):
        crc ^= data
        for _ in range(0, 8):
            if crc & 0x80 != 0:
                crc = (crc << 1) ^ 0x07
            else:
                crc <<= 1
        return crc & 0xFF

    def calculate_crc(self, bts):
        crc = 0xff
        for b in bts:
            crc = self.crc8_update(b, crc)
        return crc

    def send_and_check(self, bts):
        crc = self.calculate_crc(bts)
        bts.extend([crc])
        #print(f"Sending {bts}")
        #print(f"crc = {hex(crc)}")
        self.ser.write(bts)
        #ser.flushOutput()
        ack = self.ser.read(1)
        if ack == bytearray(STM8Programmer.ACK):
            return b'ACK'
        elif ack == bytearray(STM8Programmer.NACK):
            return b'NACK'
        return b'warning: no response'

    def bootloader_enter(self):
        # toggle reset via DTR
        self.ser.setDTR(1)
        self.ser.setRTS(1)
        sleep(0.1)
        self.ser.setDTR(0)
        sleep(0.1)
        # send payload
        req = bytearray(STM8Programmer.BOOTLOADER_INIT)
        if self.send_and_check(req) != b'ACK':
            return b'error: bootloader enter failed'
        return b'success: bootloader enter succeeded'

    def write_memory(self, file):
        # sending write memory command
        if self.send_and_check(bytearray(STM8Programmer.WRITE_MEMORY)) != b'ACK':
            return b'error: sending write memory command failed'

        # sending number of chunks
        chunks = os.path.getsize(file)
        chunks = int(math.ceil(float(chunks) / STM8Programmer.BLOCK_SIZE))
        if self.send_and_check(bytearray((chunks, ))) != b'ACK':
            return b'error: sending number of chunks failed'

        # sending chunks
        data = open(file, 'rb')
        total = 0
        with data as f:
            chunk = bytearray(f.read(STM8Programmer.BLOCK_SIZE))
            while chunk:
                total += len(chunk)
                print(total)
                # padding
                chunk.extend([0xFF] * (STM8Programmer.BLOCK_SIZE - len(chunk)))
                if self.send_and_check(chunk) != b'ACK':
                    return b'error: sending chunk failed'
                chunk = bytearray(f.read(STM8Programmer.BLOCK_SIZE))

    def read_memory(self, address = 0x8000, length = 0xff):
        # sending read memory command
        response = self.send_and_check(bytearray(STM8Programmer.READ_MEMORY))
        if response == b'NACK':
            return b'error: sending read memory command failed', b''
        elif b'warning' in response:
            return response, b''

        # sending address to read from
        address = struct.pack('>H', address)
        #print(address)
        if self.send_and_check(bytearray(address)) != b'ACK':
            return b'error: sending address to read from failed', b''
    
        # sending length
        if self.send_and_check(bytearray((length, ))) != b'ACK':
            return b'error: sending number of bytes to read failed', b''
    
        # read memory
        memory = self.ser.read(length + 1)
        crc = self.ser.read(1)
        if crc != bytes([self.calculate_crc(memory)]):
            return b'warning: bytes received, checksum wrong', memory

        return b'success:', memory

    def write_option_bytes(self, opt = 0x00):
        # sending option bytes command
        if self.send_and_check(bytearray(STM8Programmer.OPTION_WRITE)) != b'ACK':
            return b'error: sending option bytes command failed'

        # sending option byte to program
        if self.send_and_check(bytearray((opt, ))) != b'ACK':
            return b'error: sending option byte failed'

        return b'success: writing option byte successful'

    def read_option_bytes(self, length = 0x00):
        # sending option bytes command
        if self.send_and_check(bytearray(STM8Programmer.OPTION_READ)) != b'ACK':
            return b'error: sending option bytes command failed', b''

        # sending number of bytes to read
        if self.send_and_check(bytearray((length, ))) != b'ACK':
            return b'error: sending number of bytes to read failed', b''

        # read memory
        memory = self.ser.read(length + 1)
        crc = self.ser.read(1)
        if crc != bytes([self.calculate_crc(memory)]):
            return b'warning: option bytes received, checksum wrong', memory

        return b'success:', memory

    def bootloader_exec(self, args):
        response = self.bootloader_enter()
        print(response)
        if b'success' not in response:
            return

        if args.write is not None:
            response = self.write_memory(args.write)
            print(response)
            return
        
        if args.read is not None:
            response, mem = self.read_memory(args.read, args.number_of_bytes)
            print(response)
            print(mem)
            return

        if args.write_option is not None:
            response = self.write_option_bytes(args.write_option)
            print(response)
            return
        
        if args.read_option is not None:
            response, opt = self.read_option_bytes(args.read_option)
            print(response)
            print(opt)
            return

def main():
    parser = argparse.ArgumentParser(description='stm8-bootloader programmer')
    parser.add_argument('--port', '-p', required=True)
    parser.add_argument('--baud', '-b', default=115200)
    parser.add_argument('--write', '-w', required=False, help='firmware in binary format to flash to the microcontroller')
    parser.add_argument('--read', '-r', required=False, help='read bytes from address (as hex, e.g. 0x8000)', type=hex_int)
    parser.add_argument('--number-of-bytes', '-n', required=False, help='number of bytes to read (minus one)', default=0xff, type=int)
    parser.add_argument('--write-option', '-wo', required=False, help='set the value of the rop option byte (as hex, e.g. 0x00)', type=hex_int)
    parser.add_argument('--read-option', '-ro', const='0x00', required=False, help='read the value of the rop option byte', type=hex_int, nargs='?')
    args = parser.parse_args()

    programmer = STM8Programmer(args.port, args.baud)
    programmer.bootloader_exec(args)

if __name__ == "__main__":
    main()



import machine
from rp2 import asm_pio, PIO
from machine import Pin
from FastADC import FastADC

@asm_pio()
def trigger():
    wait(0, pin, 0)
    wait(1, pin, 0)
    irq(0)
    push(block)

class Test:
    def __init__(self):
        self.fastadc = FastADC()
        self.adc_samples_captured = False
        self.fastsamples = None
        self.sm0 = PIO(1).state_machine(0)
        self.pin_trigger = Pin(14, Pin.IN)
        self.led_pin = Pin(7, Pin.OUT)
        
    def arm(self):
        self.sm0.init(trigger, freq=machine.freq(), in_base=self.pin_trigger)
        self.sm0.active(1)
        
    def arm_adc(self):
        self.adc_samples_captured = False
        self.sm0.irq(handler=self.__poll_fast_adc)
 
    @micropython.native
    def __poll_fast_adc(self, sm):
        print("triggered!")
        self.fastsamples = self.fastadc.read()
        self.adc_samples_captured = True
        self.led_pin.value(not self.led_pin.value())
    
    def block(self):
        print(f"*0x50200010 = {hex(machine.mem32[0x50200010])}")
        print(f"*0x50200030 = {hex(machine.mem32[0x50200030])}")
        print(f"*0x50200038 = {hex(machine.mem32[0x50200038])}")
        print(f"*0xE000E200 = {hex(machine.mem32[0xE000E200])}")
        # wait for the statemachine to finish
        while self.sm0.rx_fifo() == 0:
            pass
        print(f"*0x50200010 = {hex(machine.mem32[0x50200010])}")
        print(f"*0x50200030 = {hex(machine.mem32[0x50200030])}")
        print(f"*0x50200038 = {hex(machine.mem32[0x50200038])}")
        print(f"*0xE000E200 = {hex(machine.mem32[0xE000E200])}")
        
    def get_adc_samples(self):
        # wait for the adc to capture samples
        while not self.adc_samples_captured:
            pass
        print(self.fastsamples)
        
reset_pin = Pin(2, Pin.OUT)
test = Test()
while True:
    test.arm()
    test.arm_adc()
    reset_pin.value(0)
    reset_pin.value(1)
    test.block()
    test.get_adc_samples()





# Copyright (C) 2024 Dr. Matthias Kesenheimer - All Rights Reserved.
# You may use, distribute and modify this code under the terms of the GPL3 license.
#
# You should have received a copy of the GPL3 license with this file.
# If not, please write to: info@faultyhardware.de.

import uarray
from machine import ADC
import Globals

@micropython.asm_thumb
def read_memory(r0): # Address
    # r0 = *(uint32_t *)r1;
    mov(r1,r0)
    ldr(r0,[r1,0])

@micropython.asm_thumb
def write_memory(r0, r1): # Address, Data
    # *(uint32_t *)r0 = r1
    str(r1,[r0,0])
    mov(r0,r1)
    
def set_memory(adress, mask):
    write_memory(adress, read_memory(adress) | mask)

@micropython.asm_thumb
def adc_get(r0, r1): # BufferAddress, Ctrl-Array
    b(Entry)
    #b(StartSample)
    
    label(GetAdc) # get one sample out of the fifo
    ldr(r2,[r1,0])
    ldrh(r2,[r2,10]) # read upper Half of FCS - something in the fifo?
    cmp(r2,0)
    bls(GetAdc)
    ldr(r2,[r1,0])
    ldr(r2,[r2,12]) # read FIFO
    bx(lr)
    
    label(Entry)
    mov(r4,r0) # BufferAdress
    #cpsid(1)
    bl(GetAdc) # empty fifo
    bl(GetAdc)
    bl(GetAdc)
    bl(GetAdc)
    bl(GetAdc)
    bl(GetAdc)
    
    label(StartSample)
    ldr(r3,[r1,8]) # Buffer Length
    label(GetNext)
    bl(GetAdc)
    strh(r2,[r4,0])
    add(r4,r4,2) # next address
    sub(r3,r3,1)
    cmp(r3,0)
    bhi(GetNext)
    
    #cpsie(1)
    label(End)

class FastADC():
    def __init__(self):
        self.configure_adc()

    def configure_adc(self, number_of_samples:int = 1024, sampling_freq:int = 500_000):
        self.number_of_samples = number_of_samples
        rGPIO26 = Globals.PADS_BANK0_BASE + 0x6c
        write_memory(rGPIO26, 1<<7) # switch off ADC0 input and output
        self.adc_buffer = uarray.array('h', range(0, number_of_samples))
        self.adc_control = uarray.array('i',[
            Globals.ADC_BASE, # BASE 0 CS, 4 Result, 8 FCS, 12 FIFO, 16 DIV S. 570
            1, # 4 Status: 0 done, 1 start
            number_of_samples # 8
            ])
        _ = ADC(0)
        #adcx.read_u16() # set up the channel; disabled line since this leads to locks
        max_sample = 48e6 # Hz, TODO: this depends on the CPU frequency!
        div = int(max_sample / sampling_freq)
        write_memory(Globals.ADC_BASE + 0x10, div<<8) # set DIV register
        set_memory(Globals.ADC_BASE + 0x08, 1) # set fifo enable
        set_memory(Globals.ADC_BASE, 1<<3) # set startmany

    def init_array(self):
        return self.adc_buffer

    def get_number_of_samples(self):
        return self.number_of_samples

    def read(self):
        adc_get(self.adc_buffer, self.adc_control)
        return self.adc_buffer


#!/usr/bin/env python3
# Copyright (C) 2024 Dr. Matthias Kesenheimer - All Rights Reserved.
# You may use, distribute and modify this code under the terms of the GPL3 license.
#
# You should have received a copy of the GPL3 license with this file.
# If not, please write to: info@faultyhardware.de.

import sys
import time
import serial
import numpy as np

from findus import Glitcher
try:
    import chipwhisperer as cw
except Exception as _:
    print("[-] Error: Library chipwhisperer not installed. Functions to interface the ChipWhisperer Pro and ChipWhisperer Husky not available.")
    print("    Install the chipwhisperer package with 'pip install chipwhisperer'")
    sys.exit(1)

class ProGlitcher(Glitcher):
    """
    Class giving access to the functions of the Chipwhisperer Pro. Derived from Glitcher class.
    Code snippet:

        from findus.ProGlitcher import ProGlitcher
        glitcher = ProGlitcher()
        glitcher.init()
        # set up database, define delay and length
        ...
        # one shot glitching
        glitcher.arm(delay, length)
        # reset target for 0.01 seconds (the rising edge on reset line triggers the glitch)
        glitcher.reset_target(0.01)
        self.glitcher.block(timeout=1)

        # read the response from the device (for example UART, SWD, etc.)
        response = ...
        # classify the response and put into database
        color = glitcher.classify(response)
        database.insert(experiment_id, delay, length, color, response)

        # reset crowbar transistors
        self.glitcher.reset_glitch()

    Methods:
        __init__: Default constructor. Does nothing in this case.
        init: Default initialization procedure.
        arm: Arm the ChipWhisperer Pro and wait for trigger condition.
        capture: Captures trace. Scope must be armed before capturing.
        block: Block the main script until trigger condition is met. Times out.
        reset_glitch: Disables and enables crowbar MOSFETs. Waits `delay` seconds in between.
        reset: Reset the target via the ChipWhisperer Pro's `RESET` output.
        power_cycle_target: Power cycle the target via the ChipWhisperer Pro `VTARGET` output.
        power_cycle_reset: Power cycle and reset the target via the ChipWhisperer Pro `RESET` and `VTARGET` output.
        reset_and_eat_it_all: Reset the target and flush the serial buffers.
        reset_wait: Reset the target and read from serial.
        set_lpglitch: Enable low-power MOSFET for glitch generation.
        set_hpglitch: Enable high-power MOSFET for glitch generation.
        rising_edge_trigger: Configure the ChipWhisperer Pro to trigger on a rising edge on the `TRIGGER` line.
        uart_trigger: Configure the ChipWhisperer Pro to trigger when a specific byte pattern is observed on the `TRIGGER` line.
        disconnect: Disconnects the ChipWhisperer Pro.
        reconnect: Disconnects and reconnects the ChipWhisperer Pro.
        reconnect_with_uart: Disconnects and reconnects the ChipWhisperer Pro. The ChipWhisperer Pro is set up for UART glitching.
        __del__: Default deconstructor. Disconnects the ChipWhisperer Pro.
    """

    def __init__(self):
        """
        Default constructor. Does nothing in this case.
        """
        self.scope = None

    def init(self, ext_power:str = None, ext_power_voltage:float = 3.3):
        """
        Default initialization procedure of the ChipWhisperer Pro. Default configuration is:

        - Set the Pro's system clock to 100 MHz.
        - Set the trigger input to rising-edge trigger on `TIO4` pin.
        - Set reset out on `nrst` pin.
        - Set serial RX on `TIO1` and TX on `TIO2` pin (necessary for UART-trigger).
        - Use the high-power crowbar MOSFET.

        Parameters:
            ext_power: Port identifier of the external power supply (RD6006). If None, target is assumed to be supplied by the voltage supplies of the ChipWhisperer Pro UFO board.
            ext_power_voltage: Supply voltage of the external power supply. Must be used in combination with `ext_power`.
        """

        try:
            self.scope = cw.scope()
        except Exception as e:
            print("[-] No ChipWhisperer found. Exiting.")
            print(f"[-] Exception: {e}")
            sys.exit(1)

        self.scope.clock.adc_src            = "clkgen_x1"
        self.scope.clock.clkgen_freq        = 100e6
        self.scope.adc.basic_mode           = "rising_edge"
        self.scope.adc.samples              = 10000
        self.scope.adc.offset               = 0
        self.scope.io.tio1                  = 'high_z'  # UART RX
        self.scope.io.tio4                  = 'high_z'  # TRIGGER
        self.scope.trigger.triggers         = 'tio4'
        self.scope.io.hs2                   = "disabled"
        self.scope.io.glitch_hp             = True
        self.scope.io.glitch_lp             = False
        # Clock asynchronous glitching
        self.scope.glitch.clk_src           = 'clkgen'
        self.scope.glitch.output            = 'enable_only'
        self.scope.glitch.trigger_src       = 'ext_single'
        if ext_power is not None:
            from findus.ExternalPowerSupply import ExternalPowerSupply
            self.power_supply = ExternalPowerSupply(port=ext_power)
            self.power_supply.set_voltage(ext_power_voltage)
            print(self.power_supply.status())
        else:
            self.power_supply = None

    def arm(self, delay:int, length:int):
        """
        Arm the ChipWhisperer Pro and wait for the trigger condition. The trigger condition can either be trigger when the reset on the target is released or when a certain pattern is observed in the serial communication.

        Parameters:
            delay: Glitch is emitted after this time. Given in nano seconds. Expect a resolution of about 10 nano seconds.
            length: Length of the glitch in nano seconds. Expect a resolution of about 10 nano seconds.
        """
        offset = delay // (int(1e9) // int(self.scope.clock.clkgen_freq))
        repeat = length // (int(1e9) // int(self.scope.clock.clkgen_freq))
        self.scope.glitch.ext_offset = offset if offset > 0 else 1
        self.scope.glitch.repeat = repeat if repeat > 0 else 1
        self.scope.arm()

    def arm_adc(self):
        """
        Arm the ADC of the ChipWhisperer Pro. This is a dummy function, as arming is done automatically with the arm() function. Still included to ensure compatibility of scripts for the Pico Glitcher.
        """
        pass

    def capture(self) -> bool:
        """
        Captures trace. Glitcher must be armed before capturing.
        Blocks until glitcher triggered (or times out), then disarms glitcher and copies data back.

        Returns:
            True if capture timed out, false if it didn't.
        Raises:
            IOError - Unknown failure.
        """
        return self.scope.capture()

    def get_adc_samples(self, timeout:float = None) -> list[int]:
        """
        Read back the captured ADC samples.
        """
        if timeout is not None:
            self.scope.adc.timeout = timeout
        samples = self.scope.get_last_trace(as_int=True).tolist()
        return samples

    def configure_adc(self, number_of_samples:int = 1024, sampling_freq:int = 100e6):
        """
        Configure the onboard ADC of the ChipWhisperer Pro.

        Parameters:
            number_of_samples: The number of samples to capture after triggering.
            sampling_freq: The sampling frequency of the ADC. `100 MSPS` is the default value for the ChipWhisperer Pro. Note, that clock generation frequency depends on this value.
        """
        self.scope.adc.samples = number_of_samples
        self.scope.clock.clkgen_freq = sampling_freq

    def block(self, timeout:float = 1):
        """
        Block until trigger condition is met. Raises an exception if times out.

        Parameters:
            timeout: Time after the block is released (not implemented yet).
        Raises:
            Timout exception.
        """
        self.scope.adc.timeout = timeout
        if self.scope.capture():
            raise Exception("Function execution timed out!")

    def reset_glitch(self, delay:float = 0.005):
        """
        Disables and enables crowbar MOSFETs. Waits `delay` seconds in between.
        Note: Up until now, only the high-power MOSFET is disabled and enabled again.

        Parameters:
            delay: Delay between disabling and re-enabling the crowbar MOSFETs.
        """
        # TODO: control hp and lp externally
        self.scope.io.glitch_hp = False
        self.scope.io.glitch_lp = False
        time.sleep(delay)
        self.scope.io.glitch_hp = True
        self.scope.io.glitch_lp = False

    def reset_target(self, reset_time:float = 0.2):
        """
        Reset the target via the ChipWhisperer Pro's `nrst` output.

        Parameters:
            reset_time: Time how long the target is held in reset.
        """
        self.scope.io.nrst = 'low'
        time.sleep(reset_time)
        self.scope.io.nrst = 'high_z'

    def reset(self, reset_time:float = 0.2):
        """
        Reset the target via the ChipWhisperer Pro's `nrst` output. This function is the same as `reset_target`. Left in for downward compatibility.

        Parameters:
            reset_time: Time how long the target is held in reset.
        """
        self.reset_target(reset_time)

    def power_cycle_target(self, power_cycle_time:float = 0.2):
        """
        Power cycle the target via the ChipWhisperer Pro's UFO board.
        If available, target is power-cycled by the external power supply RD6006.
        
        Parameters:
            power_cycle_time: Time how long the power supply is cut. If `ext_power` is defined, the external power supply (RD6006) is cycled.
        """
        if self.power_supply is not None:
            self.power_supply.power_cycle_target(power_cycle_time)
        else:
            self.scope.io.target_pwr = False
            time.sleep(power_cycle_time)
            self.scope.io.target_pwr = True

    def power_cycle_reset(self, power_cycle_time:float = 0.2):
        """
        Power cycle and reset the target via the ChipWhisperer Pro's UFO board and `nrst` output. Can also be used to define sharper trigger conditions via the `nrst` line.
        
        Parameters:
            power_cycle_time: Time how long the power supply is cut. If `ext_power` is defined, the external power supply is cycled.
        """
        if self.power_supply is not None:
            self.power_supply.disable_vtarget()
            self.scope.io.nrst = False
            time.sleep(power_cycle_time)
            self.power_supply.enable_vtarget()
            self.scope.io.nrst = "high_z"
        else:
            self.scope.io.target_pwr = False
            self.scope.io.nrst = False
            time.sleep(power_cycle_time)
            self.scope.io.target_pwr = True
            self.scope.io.nrst = "high_z"

    def reset_and_eat_it_all(self, target:serial.Serial, target_timeout:float = 0.3):
        """
        Reset the target via the ChipWhisperer Pro's `nrst` output and flush the serial buffers.

        Parameters:
            target: Serial communication object (usually defined as `target = serial.Serial(...)`).
            target_timeout: Time-out of the serial communication. After this time, reading from the serial connection is canceled and it is assumed that there is no more garbage on the line.
        """
        self.scope.io.nrst = 'low'
        target.ser.timeout = target_timeout
        target.read(4096)
        target.ser.timeout = target.timeout
        self.scope.io.nrst = 'high_z'

    def reset_wait(self, target:serial.Serial, token:bytes, reset_time:float = 0.2, debug:bool = False) -> str:
        """
        Reset the target via the ChipWhisperer Pro's `nrst` output and wait until the target responds (read from serial).

        Parameters:
            target: Serial communication object (usually defined as `target = serial.Serial(...)`).
            token: Expected response from target. Read from serial multiple times until target responds.
            reset_time:  Time how long the target is held under reset.
            debug: If `true`, more output is given.

        Returns:
            Returns the target's response.
        """
        self.scope.io.nrst = 'low'
        time.sleep(reset_time)
        self.scope.io.nrst = 'high_z'
        response = target.read(4096)
        for _ in range(0, 5):
            if token in response:
                break
            response += target.read(4096)
        if debug:
            for line in response.splitlines():
                print('\t', line.decode())
        return response

    def set_lpglitch(self):
        """
        Enable the low-power crowbar MOSFET for glitch generation.

        The glitch output is an SMA-connected output line that is normally connected to a target's power rails. If this setting is enabled, a low-powered MOSFET shorts the power-rail to ground when the glitch module's output is active.
        """
        self.scope.io.glitch_hp = False
        self.scope.io.glitch_lp = True

    def set_hpglitch(self):
        """
        Enable the high-power crowbar MOSFET for glitch generation.

        The glitch output is an SMA-connected output line that is normally connected to a target's power rails. If this setting is enabled, a high-powered MOSFET shorts the power-rail to ground when the glitch module's output is active.
        """
        self.scope.io.glitch_hp = True
        self.scope.io.glitch_lp = False

    def rising_edge_trigger(self, pin_trigger:str = "default", dead_time:float = 0, pin:str = ""):
        """
        Configure the Pico Glitcher to trigger on a rising edge on the `TRIGGER` line (`tio4` pin).
        Note: `dead_time` and `pin` have no functions here (see `PicoGlitcher.rising_edge_trigger`).

        Parameters:
            pin_trigger: The trigger input pin to use. Default is tio4.
            dead_time: Unused.
            pin: Unused.
        """
        if pin_trigger == "default":
            self.scope.trigger.triggers = 'tio4'
            self.scope.io.tio4 = 'high_z'
        else:
            self.scope.trigger.triggers = pin_trigger
            # TODO: set self.scope.io.tiox based on pin_trigger

    def uart_trigger(self, pattern:int, baudrate:int = 115200, number_of_bits:int = 8, pin_trigger:str = "default"):
        """
        Configure the ChipWhisperer Pro to trigger when a specific byte pattern is observed on the RX line (`tio1` pin).
        Note: To comply with the STM32 bootloader, this is currently configured for even parity UART.

        Parameters:
            pattern: Byte pattern that is transmitted on the serial lines to trigger on. For example `0x11`.
            baudrate: The baudrate of the serial communication.
            number_of_bits: The number of bits of the UART payload (not implemented yet, default is 8).
            pin_trigger: The trigger pin to use.
        """
        # TODO: implement the number of bits.
        # UART trigger:
        # even parity problem
        # see: https://sec-consult.com/blog/detail/secglitcher-part-1-reproducible-voltage-glitching-on-stm32-microcontrollers/
        CODE_READ = 0x80
        CODE_WRITE = 0xC0
        ADDR_DECODECFG = 57
        #ADDR_DECODEDATA = 58
        data = self.scope.decode_IO.oa.sendMessage(CODE_READ, ADDR_DECODECFG, Validate=False, maxResp=8)
        data[1] = data[1] | 0x01
        self.scope.decode_IO.oa.sendMessage(CODE_WRITE, ADDR_DECODECFG, data)
        if pin_trigger == "default":
            self.scope.trigger.triggers = 'tio1'
        else:
            self.scope.trigger.triggers = pin_trigger
        self.scope.trigger.module = 'DECODEIO'
        self.scope.decode_IO.rx_baud = baudrate
        self.scope.decode_IO.decode_type = 'USART'
        self.scope.decode_IO.trigger_pattern = [pattern]
        #self.scope.io.hs2 = "clkgen"

    def disconnect(self) -> bool:
        """
        Disconnects the ChipWhisperer Pro.

        Returns:
            True if the disconnection was successful, False otherwise.
        """
        if self.scope is not None:
            print("[+] Disconnecting ChipWhisperer Pro")
            #self.scope.io.glitch_hp = False
            #self.scope.io.glitch_lp = False
            return self.scope.dis()
        return False

    def reconnect(self, disconnect_wait:float = 0.5):
        """
        Disconnects and reconnects the ChipWhisperer Pro. The method `ProGlitcher.init()` for default initialization is called.

        Parameters:
            disconnect_wait: Time to wait during disconnects.
        """
        self.disconnect()
        time.sleep(disconnect_wait)
        self.init()

    def reconnect_with_uart(self, pattern:int, baudrate:int = 115200, number_of_bits:int = 8, disconnect_wait:float = 0.5, pin_trigger:str = "default"):
        """
        Disconnects and reconnects the ChipWhisperer Pro. The ChipWhisperer Pro is set up for UART glitching.

        Parameters:
            pattern: Byte pattern that is transmitted on the serial lines to trigger on. For example `0x11`.
            baudrate: The baudrate of the serial communication.
            number_of_bits: The number of bits of the UART payload (not implemented yet, default is 8).
            disconnect_wait: Time to wait during disconnects.
            pin_trigger: The trigger pin to use.
        """
        self.disconnect()
        time.sleep(disconnect_wait)
        self.init()
        self.uart_trigger(pattern, baudrate, number_of_bits, pin_trigger)

    def __del__(self):
        """
        Default deconstructor. Disconnects the ChipWhisperer Pro.
        """
        self.disconnect()



import rp2
from machine import Pin
import time
from FastADC import FastADC

# Define PIO program for SM0 to trigger an interrupt on button press
@rp2.asm_pio()
def button_irq():
    wrap_target()
    wait(0, pin, 0)
    wait(1, pin, 0)
    irq(block, 0)
    wrap()

# Initialize State Machine 0 for the button (GPIO 11)
button_pin = Pin(14, Pin.IN, Pin.PULL_DOWN)
sm_button = rp2.StateMachine(0, button_irq, freq=2000, in_base=button_pin)
led_pin = Pin(7, Pin.OUT)
fastadc = FastADC()
fastsamples = None
sampled = False

# Interrupt handler for the button press
def button_handler(sm):
    print("triggered")
    global sampled, fastsamples
    fastsamples = fastadc.read()
    led_pin.value(not led_pin.value())  # Toggle LED state
    sampled = True

# Attach the interrupt handler to the State Machine's IRQ
sm_button.irq(button_handler)
# Activate the state machine
sm_button.active(1)
 
# Keep the program running
while True:
    time.sleep(1)
    if sampled:
        print(fastsamples)
        sampled = False



#!/usr/bin/env python3
# Copyright (C) 2025 Dr. Matthias Kesenheimer - All Rights Reserved.
# You may use, distribute and modify this code under the terms of the GPL3 license.
#
# You should have received a copy of the GPL3 license with this file.
# If not, please write to: info@faultyhardware.de.

# programming
# > openocd -f interface/stlink.cfg -c "transport select hla_swd" -f target/stm32l0.cfg -c "init; halt; stm32l0x unlock 0; exit"
# > openocd -f interface/stlink.cfg -c "transport select hla_swd" -f target/stm32l0.cfg -c "init; halt; program rdp-downgrade-stm32l051.elf; reset run; exit;"
# > openocd -f interface/stlink.cfg -c "transport select hla_swd" -f target/stm32l0.cfg -c "init; halt; stm32l0x lock 0; exit"
# -> power cycle the target!

# reading
# >  openocd -f interface/stlink.cfg -c "transport select hla_swd" -f target/stm32l0.cfg -c "init; dump_image dump.bin 0x08000000 0x400; exit"

# debugging (install arm-none-eabi-gdb!)
# > openocd -f interface/stlink.cfg -c "transport select hla_swd" -f target/stm32l0.cfg -c "init"
# > arm-none-eabi-gdb rdp-downgrade-stm32l051.elf
# (gdb) target extended-remote localhost:3333
# (gdb) x 0x08000000
## or
# > telnet localhost 4444
# > mdw 0x08000000

# load image to ram
# > openocd -f interface/stlink.cfg -c "transport select hla_swd" -f target/stm32l0.cfg -c "init; halt; load_image rdp-downgrade-stm32l051.elf" -c "reg sp 0x20000000" -c "reg pc 0x20000004" -c "resume; exit"

# load image to ram and execute
# > openocd -f interface/stlink.cfg -c "transport select hla_swd" -f target/stm32l0.cfg -c "init; halt; load_image rdp-downgrade-stm32l051.elf" -c "reg sp 0x20002000" -c "reg pc 0x20000fa4" -c "resume" -c "exit"

# Alternatively load with gdb:
# > openocd -f interface/stlink.cfg -c "transport select hla_swd" -f target/stm32l0.cfg -c "init; halt"
# > arm-none-eabi-gdb rdp-downgrade-stm32l051.elf
# > target remote :3333
# > load rdp-downgrade-stm32l051.elf
## Start address 0x20000fa4, load size 4764
# > x $pc
## 0x20000fa4 <Reset_Handler>:	0x4685480d
# > b *0x20000fa4
# > continue
# > x $sp
## 0x20001ff8

# read RDP level
# > openocd -f interface/stlink.cfg -c "transport select hla_swd" -f target/stm32l0.cfg -c "init" -c "mdw 0x4002201c" -c "exit"


import time
import subprocess
import re
import socket

class DebugInterface():
    """
    Class to interact with the target device using OpenOCD and GDB.
    """
    def __init__(self, interface:str = "stlink", interface_config:str=None, target:str = "stm32l0", target_config:str=None, transport:str = "hla_swd", gdb_exec:str = "arm-none-eabi-gdb", adapter_serial:str = None, gdb_port = 3333, telnet_port = 4444, tcl_port = 6666):
        """
        Initializes the OpenOCD configuration with default values or custom values provided.
    
        Parameters:
            interface: The interface type, e.g., 'stlink' or 'j-link'. Defaults to 'stlink'.
            interface_config: Path to a custom interface configuration file. If None, uses the default configuration for the specified interface.
            target: The target microcontroller, e.g., 'stm32l0'. Defaults to 'stm32l0'.
            target_config: Path to a custom target configuration file. If None, uses the default configuration for the specified target.
            transport: The transport protocol, e.g., 'hla_swd'. Defaults to 'hla_swd'.
            gdb_exec: The path to the GDB executable. Defaults to 'arm-none-eabi-gdb'.
            adapter_serial: The serial number of the adapter. If None, no specific serial is used.
            gdb_port: The port number for GDB communication. Defaults to 3333.
            telnet_port: The port number for Telnet communication. Defaults to 4444.
            tcl_port: The port number for TCL communication. Defaults to 6666.
        """
        self.openocd_process = None
        self.gdb_process = None
        self.target_name = target
        self.socket = None
        self.gdb_exec = gdb_exec
        self.transport = transport
        self.adapter_serial = adapter_serial
        self.gdb_port = gdb_port
        self.telnet_port = telnet_port
        self.tcl_port = tcl_port
        if interface_config is None:
            self.interface_config = f'interface/{interface}.cfg'
        else:
            self.interface_config = interface_config
        if target_config is None:
            self.target_config = f'target/{self.target_name}.cfg'
        else:
            self.target_config = target_config

    def program_target(self, glitcher, elf_image:str = "program.elf", unlock:bool = True, rdp_level:int = 0, power_cycle_time:float = 0.1, verbose:bool = False):
        """
        Main function to program the target. Optionally unlocks the target, writes the ELF image, and adjusts the RDP level.
        
        Parameters:
            glitcher: An instance of the Glitcher class used for glitching operations.
            elf_image: The path to the ELF image file to be written to the target. Default is "program.elf".
            unlock: Whether to unlock the target before programming. Default is True.
            rdp_level: The desired RDP level (0 or 1). Default is 0.
            power_cycle_time: The time to wait between power cycles. Default is 0.1 seconds.
            verbose: Whether to print verbose output. Default is False.
        """
        if unlock:
            rdp = 0x01
            pcrop = 0x01
            while rdp != 0xaa or pcrop != 0x00:
                glitcher.reset(0.01)
                time.sleep(0.005)
                self.unlock_target(verbose=verbose)
                glitcher.power_cycle_reset(power_cycle_time)
                time.sleep(power_cycle_time)
                rdp, pcrop = self.read_rdp_and_pcrop(verbose=False)
                if verbose:
                    print(f"[+] rdp, pcrop = {hex(rdp)}, {hex(pcrop)}")
        self.write_image(elf_image=elf_image, verbose=verbose)
        rdp, pcrop = self.read_rdp_and_pcrop(verbose=False)
        if verbose:
            print(f"[+] rdp, pcrop = {hex(rdp)}, {hex(pcrop)}")
        if rdp_level == 1:
            self.lock_target(verbose=verbose)
            rdp, pcrop = self.read_rdp_and_pcrop(verbose=False)
            if verbose:
                print(f"[+] rdp, pcrop = {hex(rdp)}, {hex(pcrop)}")
            # changes in the RDP level become active after a power-cycle
            glitcher.power_cycle_reset(power_cycle_time)
            time.sleep(power_cycle_time)

    def unlock_target(self, verbose:bool = False):
        """
        Unlocks the target and removes any read-out protection.
        Attention: This will erase the targets flash!
        """
        if verbose:
            print("[+] Unlocking target...")
        args = [
            'openocd',
            '-f', self.interface_config,
            '-c', f'transport select {self.transport}',
            '-f', self.target_config,
            '-c', f'gdb_port {self.gdb_port}',
            '-c', f'telnet_port {self.telnet_port}',
            '-c', f'tcl_port {self.tcl_port}',
            '-c', f'init; halt; {self.target_name}x unlock 0; exit'
            ]
        if self.adapter_serial is not None:
            args.insert(3, '-c')
            args.insert(4, f'adapter serial {self.adapter_serial}')
        result = subprocess.run(args, text=True, capture_output=True)
        if verbose:
            print(result.stdout + result.stderr)

    def lock_target(self, verbose:bool = False):
        """
        Locks the target (activates RDP).
        """
        if verbose:
            print("[+] Locking target...")
        args = [
            'openocd',
            '-f', self.interface_config,
            '-c', f'transport select {self.transport}',
            '-f', self.target_config,
            '-c', f'gdb_port {self.gdb_port}',
            '-c', f'telnet_port {self.telnet_port}',
            '-c', f'tcl_port {self.tcl_port}',
            '-c', f'init; halt; {self.target_name}x lock 0; exit'
            ]
        if self.adapter_serial is not None:
            args.insert(3, '-c')
            args.insert(4, f'adapter serial {self.adapter_serial}')
        result = subprocess.run(args, text=True, capture_output=True)
        if verbose:
            print(result.stdout + result.stderr)

    def write_image(self, elf_image:str = "program.elf", verbose:bool = False):
        """
        Writes an ELF image to the target using OpenOCD.

        Parameters:
            elf_image: The path to the ELF image to be written. Defaults to "program.elf".
            verbose: Whether to print verbose output. Defaults to False.
        """
        if verbose:
            print("[+] Writing image to target...")
        args = [
            'openocd',
            '-f', self.interface_config,
            '-c', f'transport select {self.transport}',
            '-f', self.target_config,
            '-c', f'gdb_port {self.gdb_port}',
            '-c', f'telnet_port {self.telnet_port}',
            '-c', f'tcl_port {self.tcl_port}',
            '-c', f'init; halt; program {elf_image}; exit'
            ]
        if self.adapter_serial is not None:
            args.insert(3, '-c')
            args.insert(4, f'adapter serial {self.adapter_serial}')
        result = subprocess.run(args, text=True, capture_output=True)
        if verbose:
            print(result.stdout + result.stderr)

    def load_exec(self, elf_image:str = "program.elf", sp:int = 0x20002000, pc:int = 0x20000fa4, verbose:bool = False):
        """
        Load image to RAM and execute.
        Attention: Program must also be compiled to be compatible to run in RAM.

        Parameters:
            elf_image: The path to the ELF image file to be loaded. Default is "program.elf".
            sp: The initial stack pointer value. Default is 0x20002000.
            pc: The initial program counter value. Default is 0x20000fa4.
            verbose: If True, print the output and error from openocd. Default is False.
        """
        # openocd -f interface/stlink.cfg -c "transport select hla_swd" -f target/stm32l0.cfg -c "init; halt; load_image rdp-downgrade-STM32L0.elf" -c "reg sp 0x20002000" -c "reg pc 0x20000fa4" -c "resume" -c "exit"
        args = [
            'openocd',
            '-f', self.interface_config,
            '-c', f'transport select {self.transport}',
            '-f', self.target_config,
            '-c', f'gdb_port {self.gdb_port}',
            '-c', f'telnet_port {self.telnet_port}',
            '-c', f'tcl_port {self.tcl_port}',
            '-c', f'init; halt; load_image {elf_image}',
            '-c', f'reg sp {hex(sp)}',
            '-c', f'reg pc {hex(pc)}',
            '-c', 'resume',
            '-c', 'exit',
            ]
        if self.adapter_serial is not None:
            args.insert(3, '-c')
            args.insert(4, f'adapter serial {self.adapter_serial}')
        result = subprocess.run(args, text=True, capture_output=True)
        if verbose:
            print(result.stdout + result.stderr)

    def read_image(self, bin_image:str = "memory_dump.bin", start_addr:int = 0x08000000, length:int = 0x400, verbose:bool = False):
        """
        Reads an image from the target device's memory and saves it to a binary file.
        
        Parameters:
            bin_image: The path to the binary file where the image will be saved.
            start_addr: The starting address in memory to read from.
            length: The number of bytes to read from memory.
            verbose: If True, print the output and error messages.
        """
        args = [
            'openocd',
            '-f', self.interface_config,
            '-c', f'transport select {self.transport}',
            '-f', self.target_config,
            '-c', f'gdb_port {self.gdb_port}',
            '-c', f'telnet_port {self.telnet_port}',
            '-c', f'tcl_port {self.tcl_port}',
            '-c', f'init; dump_image {bin_image} {hex(start_addr)} {hex(length)}; exit'
            ]
        if self.adapter_serial is not None:
            args.insert(3, '-c')
            args.insert(4, f'adapter serial {self.adapter_serial}')
        result = subprocess.run(args, text=True, capture_output=True)
        if verbose:
            print(result.stdout + result.stderr)

    def test_connection(self):
        """
        Tests the connection to the target device by resetting it.
        """
        args = [
            'openocd',
            '-f', self.interface_config,
            '-c', f'transport select {self.transport}',
            '-f', self.target_config,
            '-c', f'gdb_port {self.gdb_port}',
            '-c', f'telnet_port {self.telnet_port}',
            '-c', f'tcl_port {self.tcl_port}',
            '-c', 'init; reset run; exit'
            ]
        if self.adapter_serial is not None:
            args.insert(3, '-c')
            args.insert(4, f'adapter serial {self.adapter_serial}')
        result = subprocess.run(args, text=True, capture_output=True)
        print(result.stdout + result.stderr)

    def read_address(self, address):
        """
        Read memory content from a specific address using OpenOCD.
        
        Parameters:
            address: The memory address to read from.
        
        Returns:
            A tuple containing the extracted memory content and the raw response.
        """
        args = [
            'openocd',
            '-f', self.interface_config,
            '-c', f'transport select {self.transport}',
            '-f', self.target_config,
            '-c', f'gdb_port {self.gdb_port}',
            '-c', f'telnet_port {self.telnet_port}',
            '-c', f'tcl_port {self.tcl_port}',
            '-c', 'init',
            '-c', f'mdw {hex(address)}',
            '-c', 'exit'
            ]
        if self.adapter_serial is not None:
            args.insert(3, '-c')
            args.insert(4, f'adapter serial {self.adapter_serial}')
        result = subprocess.run(args, text=True, capture_output=True)
        response = result.stdout + result.stderr
        return self.extract_memory_content(response=response, address=address), response

    def read_option_bytes(self):
        """
        Reads the option bytes from the memory address 0x4002201c.

        Returns:
            A tuple containing the memory value and the response from the read operation.
            If the memory value is None, returns (0x00, response).
        """
        mem, response = self.read_address(0x4002201c)
        if mem is not None:
            return mem, response
        return 0x00, response

    def read_pcrop(self):
        """
        Reads the PCROP (Programmable Code Protection Region) value from the option bytes.

        Returns:
            The PCROP value.
        """
        optbytes, _ = self.read_option_bytes()
        pcrop = (optbytes & 0x100) >> 8
        return pcrop

    def read_rdp(self):
        """
        Reads the RDP (Read Protection) value from the option bytes.

        Returns:
            int: The RDP value.
        """
        optbytes, _ = self.read_option_bytes()
        rdp = (optbytes & 0xff)
        return rdp

    def read_rdp_and_pcrop(self, verbose:bool = False):
        """
        Reads both RDP and PCROP values from the option bytes.

        Parameters:
            verbose: If True, prints the response from the read operation. Defaults to False.

        Returns:
            A tuple containing the RDP and PCROP values.
        """
        optbytes, response = self.read_option_bytes()
        if verbose:
            print(response)
        pcrop = (optbytes & 0x100) >> 8
        rdp = (optbytes & 0xff)
        return rdp, pcrop

    def kill_process(self, port:int, verbose=False):
        """
        Kill the process that is using the specified port.
        
        Parameters:
            port: The port number to check for a process.
            verbose: If True, print detailed output.
        """
        # Find process using the port
        cmd = ["lsof", "-i", f":{port}"]
        #pattern = r"\b(\d+)\b"  # Regex for PID
        pattern = r"openocd\s+(\d+)"  # Regex for PID
        # Run command and search for PID
        # trunk-ignore(bandit/B603)
        result = subprocess.run(cmd, capture_output=True, text=True)
        if verbose:
            print(result)
        matches = re.findall(pattern, result.stdout)
        if matches:
            pid = matches[-1]  # Get the last match (PID)
            print(f"[+] Found process {pid} using port {port}")
            # Kill the process
            try:
                # trunk-ignore(bandit/B607)
                # trunk-ignore(bandit/B603)
                subprocess.run(["kill", "-9", pid], check=True)
                print(f"[+] Process {pid} killed successfully.")
            except Exception as e:
                print(f"[-] Error killing process: {e}")

    def attach(self, delay=0.1):
        """
        Attaches to the OpenOCD server.

        Parameters:
            delay: Delay in seconds before checking if OpenOCD is running.
        """
        # check if there is a dangling process that would interfere with openocd
        self.kill_process(self.gdb_port)
        args = [
            'openocd',
            '-f', self.interface_config,
            '-c', f'transport select {self.transport}',
            '-f', self.target_config,
            '-c', f'gdb_port {self.gdb_port}',
            '-c', f'telnet_port {self.telnet_port}',
            '-c', f'tcl_port {self.tcl_port}',
            '-c', 'init'
            ]
        if self.adapter_serial is not None:
            args.insert(3, '-c')
            args.insert(4, f'adapter serial {self.adapter_serial}')
        try:
            self.openocd_process = subprocess.Popen(
                args,
                stdout=subprocess.PIPE,
                stderr=subprocess.PIPE,
                text=True,
                start_new_session=True,
            )
            time.sleep(delay)
            if self.openocd_process.poll() is not None:
                out, err = self.openocd_process.communicate()
                raise RuntimeError(f"OpenOCD failed:\n{out}\n{err}")
        except subprocess.TimeoutExpired:
            pass

    def mi_wait_done(self, timeout=1.0, ok_prefixes=("^done", "^running", "^connected"), verbose=False):
        """
        Wait for a GDB MI command to complete with a timeout.
        
        Parameters:
            timeout: Maximum time to wait for the command to complete (default: 1.0 seconds)
            ok_prefixes: Prefixes indicating a successful completion (default: ("^done", "^running", "^connected"))
            verbose: Whether to print each line received (default: False)
        
        Raises:
            TimeoutError: If the command times out
            RuntimeError: If the GDB process exits unexpectedly or encounters an error
        """
        start = time.time()
        while True:
            if time.time() - start > timeout:
                raise TimeoutError("GDB MI command timed out")
            line = self.gdb_process.stdout.readline()
            if not line:
                raise RuntimeError("GDB exited unexpectedly")
            line = line.strip()
            if verbose:
                print(line)
            # Ignore CLI prompt and async noise
            if not line or line == "(gdb)":
                continue
            if line[0] in "*=~&":
                continue
            # Result records
            if line.startswith("^error"):
                raise RuntimeError(f"GDB MI error: {line}")
            for ok in ok_prefixes:
                if line.startswith(ok):
                    return
        
    def mi(self, cmd, verbose=False):
        """
        Sends a GDB command through the GDB process.
    
        Parameters:
            cmd: The GDB command to send.
            verbose: If True, prints the command before sending it. Defaults to False.
        """
        if verbose:
            print(f">>> {cmd}")
        self.gdb_process.stdin.write(cmd + "\n")
        self.gdb_process.stdin.flush()

    def gdb_load_exec(self, elf_image:str="program.elf", timeout=0.3, verbose=False):
        """
        Load and execute an ELF image using GDB.

        Parameters:
            elf_image: Path to the ELF image to be loaded and executed.
            timeout: Timeout for GDB operations.
            verbose: Whether to print verbose output.
        """
        self.gdb_process = subprocess.Popen(
            [self.gdb_exec, '--interpreter=mi2', elf_image],
            stdin=subprocess.PIPE,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            text=True,
            encoding="utf-8",
            errors="replace",
            bufsize=1,  # line-buffered
        )
        self.mi(f"-file-exec-and-symbols {elf_image}", verbose=verbose)
        self.mi_wait_done(timeout=timeout, verbose=verbose)
        self.mi(f"-target-select extended-remote localhost:{self.gdb_port}", verbose=verbose)
        self.mi_wait_done(timeout=timeout, ok_prefixes=("^connected",), verbose=verbose)
        self.mi("-target-download", verbose=verbose)
        self.mi_wait_done(timeout=timeout, ok_prefixes=("^done",), verbose=verbose)
        self.mi("-exec-continue", verbose=verbose)
        self.mi_wait_done(timeout=timeout, ok_prefixes=("^running",), verbose=verbose)

    def gdb_interrupt_disconnect(self, timeout=0.3, halt_target=True, verbose=False):
        """
        Disconnects the GDB target and exits the GDB process.

        Parameters:
            timeout: The maximum time to wait for GDB operations to complete.
            halt_target: Whether to halt the target before disconnecting.
            verbose: Whether to print verbose output.
        """
        if halt_target:
            self.mi("-exec-interrupt", verbose=verbose)
            self.mi_wait_done(timeout=timeout, ok_prefixes=("^done",), verbose=verbose)
            self.mi("-target-disconnect", verbose=verbose)
            self.mi_wait_done(timeout=timeout, ok_prefixes=("^done",), verbose=verbose)
            self.mi("-gdb-exit", verbose=verbose)
            self.mi_wait_done(timeout=timeout, ok_prefixes=("^exit",), verbose=verbose)
        self.gdb_process.terminate()
        self.gdb_process.wait(timeout=timeout)

    def detach(self):
        """
        Detaches from the current debugging session by terminating and closing the necessary processes and socket.
        """
        if self.openocd_process is not None:
            self.openocd_process.terminate()
            self.openocd_process.wait()
            self.openocd_process = None
        if self.socket is not None:
            self.socket.close()
            self.socket = None
        if self.gdb_process is not None:
            self.gdb_process.terminate()
            self.gdb_process = None

    def telnet_init(self):
        """
        Establishes a connection to the OpenOCD telnet server and receives initial messages.
        """
        # generate a connection to the openocd telnet server
        self.socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        self.socket.settimeout(1)
        self.socket.connect(('localhost', {self.telnet_port}))
        # receive start messages
        time.sleep(0.1)
        self.socket.recv(4096)

    def telnet_interact(self, command:str, wait_time:float = 0.01, verbose:bool = False) -> str:
        """
        Sends a command via Telnet and receives the response.

        Parameters:
            command: The command to send.
            wait_time: Time to wait between sending the command and receiving the response. Defaults to 0.01.
            verbose: Whether to print the response. Defaults to False.

        Returns:
            The response received from the Telnet server.
        """
        if self.openocd_process is None or self.socket is None:
            self.telnet_init()
        command += "\n"
        self.socket.sendall(command.encode("utf-8"))
        time.sleep(wait_time)
        response = self.socket.recv(4096)
        if verbose:
            print(response.decode("utf-8"))
        return response.decode("utf-8")

    def telnet_read_address(self, address:int) -> tuple:
        """
        Reads the memory content at a specified address using Telnet.

        Parameters:
            address: The memory address to read.

        Returns:
            A tuple containing the memory content and the response from the Telnet command.
        """
        command = f"mdw {hex(address)}"
        response = self.telnet_interact(command)
        if "Previous state query failed, trying to reconnect" in response:
            response += self.telnet_interact(command)
        return self.extract_memory_content(response=response, address=address), response

    def telnet_read_image(self, bin_image:str = "memory_dump.bin", start_addr:int = 0x08000000, length:int = 0x400, verbose:bool = False):
        """
        Reads an image from a specified memory address using a telnet connection.
    
        Parameters:
            bin_image: The filename where the dumped image will be saved.
            start_addr: The starting memory address to read from.
            length: The length of the data to read.
            verbose: If True, prints the response from the telnet server.
        """
        command = f"init; dump_image {bin_image} {hex(start_addr)} {hex(length)}; exit"
        response = self.telnet_interact(command)
        if verbose:
            print(response)

    def extract_memory_content(self, response:str, address:int = 0x00) -> int | None:
        """
        Extracts memory content from the response.
        
        Parameters:
            response: The response string to extract memory content from.
            address: The memory address to search for. Defaults to 0x00.
        
        Returns:
            The extracted memory content if found, otherwise None.
        """
        if "Error: Failed to read memory at" not in response:
            match = re.search(fr'{hex(address)[2:]}:\s*([0-9A-Fa-f]+)', response)
            if match:
                return int(match.group(1), 16)
            else:
                return None

    def characterize(self, response:str, mem:int) -> bytes:
        """
        Characterizes the response based on the memory content.
        
        Parameters:
            response: The response string to characterize.
            mem: The memory content to check.

        Returns:
            bytes: A byte string representing the characterization.
        """
        # possibly ok
        if mem is not None:
            if mem != 0x00:
                return b'success: RDP inactive'
            else:
                return b'ok: read zero'
        # Error: no connection
        elif "Error: init mode failed (unable to connect to the target)" in response:
            return b'error: no connection'
        # Warning: Polling failed
        elif "Polling target" in response:
            return b'warning: polling failed'
        # Warning: Device lockup
        if "clearing lockup after double fault" in response:
            return b'warning: lock-up'
        # Warning: else
        elif "Warning" in response:
            return b'warning: default warning'
        # expected state
        elif "Error: Failed to read memory at" in response:
            return b'expected: RDP active'
        # no response
        return b'error: no response'

    def __del__(self):
        """
        Destructor method called when the object is about to be garbage collected.
        Detaches the debugger by printing a message and invoking the detach method.
        """
        print("[+] Detaching debugger.")
        self.detach()


# Copyright (C) 2024 Dr. Matthias Kesenheimer - All Rights Reserved.
# You may use, distribute and modify this code under the terms of the GPL3 license.
#
# You should have received a copy of the GPL3 license with this file.
# If not, please write to: info@faultyhardware.de.

import ujson
from rp2 import PIO

# load config
with open("config.json", "r") as file:
    config = ujson.load(file)

if config["hardware_version"][0] == 1:
    # Trigger 1 (without level shifter)
    ALT_TRIGGER = 18
    # Trigger 2 (with level shifter)
    TRIGGER = 15
    VTARGET_OC = 21
    VTARGET_EN = 20
    RESET = 0
    GLITCH_EN = 1
    HP_GLITCH = 16
    LP_GLITCH = 17
    # added as dummy variables to fix undefined variable error
    MUX0_PIO_INIT = None
    MUX1_PIO_INIT = None
    MUX_PIO_INIT = 0b00
elif config["hardware_version"][0] == 2 or config["hardware_version"][0] == 3:
    TRIGGER = 14
    # alternative trigger on EXT1
    ALT_TRIGGER = 11
    VTARGET_EN = 22
    RESET = 2
    GLITCH_EN = 3
    HP_GLITCH = 12
    HP_GLITCH_LED = 8
    LP_GLITCH = 13
    LP_GLITCH_LED = 7
    MUX0 = 1
    MUX1 = 0
    EXT1 = 11
    EXT2 = 10
    if config["mux_vinit"] == "GND":
        MUX1_INIT = 1
        MUX0_INIT = 1
        MUX1_PIO_INIT = PIO.OUT_HIGH
        MUX0_PIO_INIT = PIO.OUT_HIGH
        MUX_PIO_INIT = 0b11
    elif config["mux_vinit"] == "VI1" or config["mux_vinit"] == "VCC":
        MUX1_INIT = 0
        MUX0_INIT = 0
        MUX1_PIO_INIT = PIO.OUT_LOW
        MUX0_PIO_INIT = PIO.OUT_LOW
        MUX_PIO_INIT = 0b00
    elif config["mux_vinit"] == "1.8":
        MUX1_INIT = 1
        MUX0_INIT = 0
        MUX1_PIO_INIT = PIO.OUT_HIGH
        MUX0_PIO_INIT = PIO.OUT_LOW
        MUX_PIO_INIT = 0b01
    else: # 3.3 or VI2
        MUX1_INIT = 0
        MUX0_INIT = 1
        MUX1_PIO_INIT = PIO.OUT_LOW
        MUX0_PIO_INIT = PIO.OUT_HIGH
        MUX_PIO_INIT = 0b10

# FastADC config
if config["hardware_version"][0] == 2:
    PADS_BANK0_BASE = 0x4001c000 # S. 301, rp2040-datahseet.pdf
    ADC_BASE = 0x4004c000 # S. 25, rp2040-datahseet.pdf
elif config["hardware_version"][0] == 3:
    PADS_BANK0_BASE = 0x40038000 # S. 786, rp2350-datahseet.pdf
    ADC_BASE = 0x400a0000 # S. 32, rp2350-datahseet.pdf


#!/usr/bin/env python3
# Copyright (C) 2025 Bedri Zija - All Rights Reserved.
# You may use, distribute and modify this code under the terms of the GPL3 license.
#
# You should have received a copy of the GPL3 license with this file.
# If not, please write to: info@faultyhardware.de.

import numpy as np
import matplotlib.pyplot as plt
from matplotlib.widgets import Button
from scipy.interpolate import PchipInterpolator
try:
    import tkinter as tk
    from tkinter.scrolledtext import ScrolledText
except Exception as _:
    pass

class InteractivePchipEditor:
    def __init__(self):
        """
        Initialize the plot with control points, curve, and interactive elements.
        Sets up the figure, axes, data points, annotations, and event handlers.
        """
        self.xpoints = [0, 100, 200, 300, 400, 500, 600, 700, 800, 900, 1050]
        self.ypoints = [3.3, 2.7, 2.7, 0.0, 2.7, 0.0, 2.7, 0.0, 2.7, 3.0, 3.3]

        self.fig, self.ax = plt.subplots()
        plt.subplots_adjust(bottom=0.3)
        self.selected_point = None
        self.annotations = []  # To store annotation objects for the points

        self.curve_line, = self.ax.plot([], [], label="PCHIP Curve", color="blue")
        self.point_scatter = self.ax.scatter(self.xpoints, self.ypoints, color="red", s=50, label="Control Points", picker=True)

        # Set axis limits and finer ticks
        self.ax.set_xlim(-50, 1550)
        self.ax.set_ylim(-0.2, 3.5)
        self.ax.set_xticks(np.arange(0, 1550, 50))  # X-axis gridlines at 50 ns
        self.ax.set_yticks(np.arange(0, 3.4, 0.1))  # Y-axis gridlines at 0.1 V

        self.ax.set_xlabel("Time (ns)")
        self.ax.set_ylabel("Voltage (V)")
        self.ax.legend()
        self.ax.grid(True)

        save_ax = plt.axes([0.8, 0.05, 0.1, 0.075])
        self.save_button = Button(save_ax, 'Save')
        self.save_button.on_clicked(self.save_points)

        self.update_curve()
        self.update_annotations()  # Initialize annotations
        self.cid_click = self.fig.canvas.mpl_connect('button_press_event', self.on_click)
        self.cid_release = self.fig.canvas.mpl_connect('button_release_event', self.on_release)
        self.cid_motion = self.fig.canvas.mpl_connect('motion_notify_event', self.on_motion)

    def update_curve(self):
        """
        Updates the curve based on the stored points using PCHIP interpolation.
        """
        interpolator = PchipInterpolator(self.xpoints, self.ypoints)
        x_smooth = np.linspace(min(self.xpoints), max(self.xpoints), 500)
        y_smooth = interpolator(x_smooth)
        self.curve_line.set_data(x_smooth, y_smooth)
        self.point_scatter.set_offsets(np.c_[self.xpoints, self.ypoints])
        self.fig.canvas.draw_idle()

    def update_annotations(self):
        """
        Clears all existing annotations associated with the object and resets the annotations list.
        """
        # Clear existing annotations
        for annotation in self.annotations:
            annotation.remove()
        self.annotations = []

        # Add new annotations for each point
        for x, y in zip(self.xpoints, self.ypoints):
            annotation = self.ax.annotate(
                f"{x:.0f}ns, {y:.1f}V",
                (x, y),
                textcoords="offset points",
                xytext=(5, 5),  # Offset to prevent overlap
                ha="center",
                fontsize=8,
                color="black",
                bbox=dict(boxstyle="round,pad=0.3", edgecolor="gray", facecolor="white", alpha=0.8),
            )
            self.annotations.append(annotation)
        self.fig.canvas.draw_idle()

    def on_click(self, event):
        """
        Handles mouse click events to select a point from the scatter plot.
        Updates self.selected_point if a point is clicked.
        
        Parameters:
            event: The mouse click event object.
        """
        if event.inaxes != self.ax:
            return
        contains, index = self.point_scatter.contains(event)
        if contains:
            self.selected_point = index["ind"][0]

    def on_release(self, event):
        """
        Handle the release event by deselecting the current point.
        Resets the selected_point attribute to None when the mouse button is released.

        Parameters:
            event: The mouse motion event containing xdata and ydata coordinates.
        """
        self.selected_point = None

    def on_motion(self, event):
        """
        Handles mouse motion events to adjust the position of a selected point.
        Modifies the selected point's coordinates, keeps xpoints sorted, and updates
        the curve and annotations on the plot.
        
        Parameters:
            event: The mouse motion event containing xdata and ydata coordinates.
        """
        if self.selected_point is None or event.inaxes != self.ax:
            return

        # Update point position
        new_x = np.clip(event.xdata, 0, 1500)
        new_y = np.clip(event.ydata, 0, 3.3)
        self.xpoints[self.selected_point] = new_x
        self.ypoints[self.selected_point] = new_y

        # Keep xpoints sorted
        self.xpoints, self.ypoints = zip(*sorted(zip(self.xpoints, self.ypoints)))
        self.xpoints, self.ypoints = list(self.xpoints), list(self.ypoints)

        # Update the curve and annotations
        self.update_curve()
        self.update_annotations()

    def save_points(self, event):
        """
        Formats and saves the x and y points into a string representation.
        Attempts to create a text window with the formatted points. If an error occurs,
        the output is printed to the console instead.
    
        Parameters:
            event: The event that triggered this method (not used in the current implementation).
        """
        xpoints_str = ", ".join(f"{x:.0f}" for x in self.xpoints)
        ypoints_str = ", ".join(f"{y:.1f}" for y in self.ypoints)
        output = f"xpoints = [{xpoints_str}]\nypoints = [{ypoints_str}]"
        try:
            self.create_text_window(output)
        except Exception as _:
            print(output)

    def get_points(self):
        """
        Converts each element in self.xpoints and self.ypoints to float.

        Returns:
            Returns the x and y points as lists of floats.
        """
        return [float(x) for x in self.xpoints], [float(x) for x in self.ypoints]

    def show(self, block=True):
        """
        Displays the current plot and pauses briefly to ensure it is updated properly.
    
        Parameters:
            block: If True, the display call blocks until the window is closed. If False, the display call returns immediately.
        """
        plt.show(block=block)
        plt.pause(0.001)

    def create_text_window(self, output):
        """
        Creates a new topmost window displaying the provided output text.
        The window features a scrollable text area with a fixed font and size.
        The text area is set to be editable to allow copying of content.

        Parameters:
            output: The text content to be displayed in the window.
        """
        self.root = tk.Tk()
        self.root.title("Saved Points")
        self.text_output = ScrolledText(self.root, wrap=tk.WORD, width=80, height=5, font=("Courier", 10))
        self.text_output.pack(fill=tk.BOTH, expand=True)
        self.text_output.insert(tk.END, output)
        self.text_output.configure(state='normal')  # Enable copying
        self.root.attributes("-topmost", True)  # Keep window on top
        self.root.mainloop()

def main():
    editor = InteractivePchipEditor()
    editor.show()
    editor.get_points()

if __name__ == "__main__":
    main()



# Copyright (C) 2024 Dr. Matthias Kesenheimer - All Rights Reserved.
# You may use, distribute and modify this code under the terms of the GPL3 license.
#
# You should have received a copy of the GPL3 license with this file.
# If not, please write to: info@faultyhardware.de.

from rp2 import asm_pio, PIO
import Globals

@asm_pio(set_init=(PIO.OUT_LOW), sideset_init=(PIO.OUT_LOW))
def glitch():
    # block until delay received
    pull(block)
    mov(x, osr)
    # block until length received
    pull(block)
    mov(y, osr)

    # wait for trigger condition
    # enable pin_glitch_en
    wait(1, irq, 1).side(0b1)
    # set irq0 which starts the adc callback (if enabled)
    irq(0)

    # wait delay
    label("delay_loop")
    jmp(x_dec, "delay_loop")

    # emit glitch at base pin
    set(pins, 0b1)
    label("length_loop")       
    jmp(y_dec, "length_loop")

    # stop glitch and disable pin_glitch_en
    set(pins, 0b0).side(0b0)

    # tell execution finished (fills the sm's fifo buffer)
    push(block)

@asm_pio(set_init=(PIO.OUT_LOW), sideset_init=(PIO.OUT_LOW), out_shiftdir=PIO.SHIFT_RIGHT)
def glitch_burst():
    # block until delay received
    pull(block)
    mov(x, osr)
    # block until pulse config (length and delay between pulses) received, backup in isr
    pull(block)
    mov(isr, osr)
    # block until number of pulses received
    pull(block)
    mov(y, osr)

    # wait for trigger condition
    # enable pin_glitch_en
    wait(1, irq, 1).side(0b1)
    # set irq0 which starts the adc callback (if enabled)
    irq(0)

    # delay until first glitch
    label("delay_loop")
    jmp(x_dec, "delay_loop")

    # emit a certain number of glitches
    label("burst_loop")

    # copy config into osr
    mov(osr, isr)
    out(x, 16) # length = OSR >> 16

    # emit glitch
    set(pins, 0b1)
    label("length_loop")       
    jmp(x_dec, "length_loop")
    set(pins, 0b0)

    # wait until next glitch
    out(x, 16) # delay = OSR >> 16
    label("delay2_loop")       
    jmp(x_dec, "delay2_loop")

    jmp(y_dec, "burst_loop")

    # reset glitch_en
    set(pins, 0b0).side(0b0)

    # tell execution finished (fills the sm's fifo buffer)
    push(block)

@asm_pio(set_init=(PIO.OUT_LOW), sideset_init=(PIO.OUT_LOW), out_shiftdir=PIO.SHIFT_RIGHT)
def glitch_multiple():
    # block until number of pulses received
    pull(block)
    mov(y, osr)
    
    # wait for trigger condition
    # enable pin_glitch_en
    wait(1, irq, 1).side(0b1)
    # set irq0 which starts the adc callback (if enabled)
    irq(0)

    # emit multiple glitches given by each config
    label("glitch_loop")

    # block until first config is received
    pull(block)
    out(x, 22) # delay = OSR >> 22

    # delay until first glitch
    label("delay_loop")
    jmp(x_dec, "delay_loop")

    # get the length
    out(x, 10) # length = OSR >> 10

    # emit glitch at base pin
    set(pins, 0b1)
    label("length_loop")       
    jmp(x_dec, "length_loop")
    set(pins, 0b0)

    jmp(y_dec, "glitch_loop")

    # stop glitch and disable pin_glitch_en
    set(pins, 0b0).side(0b0)

    # tell execution finished (fills the sm's fifo buffer)
    push(block)

@asm_pio(set_init=(PIO.OUT_HIGH), sideset_init=(PIO.OUT_LOW))
def pulse_shaping():
    # block until delay received
    pull(block)
    mov(x, osr)
    # block until total length of pulse received
    pull(block)
    mov(y, osr)

    # wait for trigger condition
    # enable pin_glitch_en
    wait(1, irq, 1).side(0b1)
    # set irq0 which starts the adc callback (if enabled)
    irq(0)

    # wait delay
    label("delay_loop")
    jmp(x_dec, "delay_loop")

    # set trigger pin low (start pulse generator)
    set(pins, 0b0)
    label("length_loop")
    jmp(y_dec, "length_loop")

    # stop pulse and disable pin_glitch_en
    set(pins, 0b1).side(0b0)

    # tell execution finished (fills the sm's fifo buffer)
    push(block)

@asm_pio(set_init=(Globals.MUX1_PIO_INIT, Globals.MUX0_PIO_INIT), out_init=(Globals.MUX1_PIO_INIT, Globals.MUX0_PIO_INIT), sideset_init=(PIO.OUT_LOW), out_shiftdir=PIO.SHIFT_RIGHT)
def multiplex(MUX_PIO_INIT=Globals.MUX_PIO_INIT):
    # block until delay received
    pull(block)
    mov(x, osr)

    # wait for trigger condition
    # enable pin_glitch_en
    wait(1, irq, 1).side(0b1)
    # set irq0 which starts the adc callback (if enabled)
    irq(0)

    # wait delay
    label("delay_loop")
    jmp(x_dec, "delay_loop")

    # do the following two times
    set(x, 1)
    label("for_loop")

    # block until config received
    pull(block)
    out(pins, 2)
    out(y, 14) # t = OSR >> 14
    label("t1_loop")
    jmp(y_dec, "t1_loop")

    out(pins, 2)
    out(y, 14) # t = OSR >> 14
    label("t2_loop")
    jmp(y_dec, "t2_loop")

    # for loop
    jmp(x_dec, "for_loop")

    # reset and disable pin_glitch_en
    set(pins, MUX_PIO_INIT).side(0b0)

    # tell execution finished (fills the sm's fifo buffer)
    push(block)

@asm_pio(set_init=(PIO.OUT_LOW, PIO.OUT_LOW), out_init=(PIO.OUT_LOW, PIO.OUT_LOW), sideset_init=(PIO.OUT_LOW), out_shiftdir=PIO.SHIFT_RIGHT)
def multiplex_vin1(MUX_PIO_INIT=0b00):
    # block until delay received
    pull(block)
    mov(x, osr)

    # wait for trigger condition
    # enable pin_glitch_en
    wait(1, irq, 1).side(0b1)
    # set irq0 which starts the adc callback (if enabled)
    irq(0)

    # wait delay
    label("delay_loop")
    jmp(x_dec, "delay_loop")

    # do the following two times
    set(x, 1)
    label("for_loop")

    # block until config received
    pull(block)
    out(pins, 2)
    out(y, 14) # t = OSR >> 14
    label("t1_loop")
    jmp(y_dec, "t1_loop")

    out(pins, 2)
    out(y, 14) # t = OSR >> 14
    label("t2_loop")
    jmp(y_dec, "t2_loop")

    # for loop
    jmp(x_dec, "for_loop")

    # reset and disable pin_glitch_en
    set(pins, MUX_PIO_INIT).side(0b0)

    # tell execution finished (fills the sm's fifo buffer)
    push(block)

@asm_pio(set_init=(PIO.OUT_LOW, PIO.OUT_HIGH), out_init=(PIO.OUT_LOW, PIO.OUT_HIGH), sideset_init=(PIO.OUT_LOW), out_shiftdir=PIO.SHIFT_RIGHT)
def multiplex_vin2(MUX_PIO_INIT=0b10):
    # block until delay received
    pull(block)
    mov(x, osr)

    # wait for trigger condition
    # enable pin_glitch_en
    wait(1, irq, 1).side(0b1)
    # set irq0 which starts the adc callback (if enabled)
    irq(0)

    # wait delay
    label("delay_loop")
    jmp(x_dec, "delay_loop")

    # do the following two times
    set(x, 1)
    label("for_loop")

    # block until config received
    pull(block)
    out(pins, 2)
    out(y, 14) # t = OSR >> 14
    label("t1_loop")
    jmp(y_dec, "t1_loop")

    out(pins, 2)
    out(y, 14) # t = OSR >> 14
    label("t2_loop")
    jmp(y_dec, "t2_loop")

    # for loop
    jmp(x_dec, "for_loop")

    # reset and disable pin_glitch_en
    set(pins, MUX_PIO_INIT).side(0b0)

    # tell execution finished (fills the sm's fifo buffer)
    push(block)

@asm_pio()
def tio_trigger_with_dead_time_rising_edge():
    # wait for irq in block_rising_condition or block_falling_condition state machine (dead time)
    wait(1, irq, 2)

    # wait for rising edge on trigger pin
    wait(0, pin, 0)
    wait(1, pin, 0)

    # tell observed trigger
    irq(block, 1)
    push(block)

@asm_pio()
def tio_trigger_with_dead_time_falling_edge():
    # wait for irq in block_rising_condition or block_falling_condition state machine (dead time)
    wait(1, irq, 2)

    # wait for falling edge on trigger pin
    wait(1, pin, 0)
    wait(0, pin, 0)

    # tell observed trigger
    irq(block, 1)
    push(block)

@asm_pio()
def edge_trigger_rising_edge():
    # block until number of edges received
    pull(block)
    mov(x, osr)

    # count the rising edges on trigger pin
    label("edge_count_loop")

    # wait for rising edge on trigger pin
    wait(0, pin, 0)
    wait(1, pin, 0)

    # decrease x and jump to the beginning of the loop
    jmp(x_dec, "edge_count_loop")

    # tell observed trigger
    irq(block, 1)
    push(block)

@asm_pio()
def edge_trigger_falling_edge():
    # block until number of edges received
    pull(block)
    mov(x, osr)

    # count the falling edges on trigger pin
    label("edge_count_loop")

    # wait for falling edge on trigger pin
    wait(1, pin, 0)
    wait(0, pin, 0)

    # decrease x and jump to the beginning of the loop
    jmp(x_dec, "edge_count_loop")

    # tell observed trigger
    irq(block, 1)
    push(block)

@asm_pio(in_shiftdir=PIO.SHIFT_RIGHT, out_shiftdir=PIO.SHIFT_LEFT)
def uart_trigger():
    # block until pattern received
    pull(block)
    mov(x, osr)
    # block until number of bits (self.number_of_bits - 1) received, store in osr
    pull(block)

    label("start")
    mov(isr, null)
    # Wait for start bit
    wait(0, pin, 0)
    # Preload bit counter, delay until eye of first data bit
    #set(y, BITS - 1) [10]
    mov(y, osr) [10]

    # Loop 9 times
    label("bitloop")
    # Sample data, shift sampled data into ISR
    in_(pins, 1)
    # Each iteration is 8 cycles
    jmp(y_dec, "bitloop") [6]

    # compare received data with supplied pattern
    mov(y, isr)
    jmp(x_not_y, "start")

    # if received data matches pattern, set the irq and activate the glitch
    irq(block, 1)
    push(block)

    # wrap around # TODO: ist hier ein wrap überhaupt notwendig?
    jmp("start")

@asm_pio()
def block_rising_condition():
    # block until dead time parameters received
    pull(block)
    mov(x, osr)

    # wait for rising edge condition
    wait(1, pin, 0)

    # wait dead time
    label("delay_loop")
    jmp(x_dec, "delay_loop")

    # tell execution finished
    irq(block, 2)
    push(block)

@asm_pio()
def block_falling_condition():
    # block until dead time parameters received
    pull(block)
    mov(x, osr)

    # wait for falling edge condition
    wait(0, pin, 0)

    # wait dead time
    label("delay_loop")
    jmp(x_dec, "delay_loop")

    # tell execution finished
    irq(block, 2)
    push(block)


# Copyright (C) 2024 Dr. Matthias Kesenheimer - All Rights Reserved.
# You may use, distribute and modify this code under the terms of the GPL3 license.
#
# You should have received a copy of the GPL3 license with this file.
# If not, please write to: info@faultyhardware.de.

# Fast-Cubic-Spline-Python provides an implementation of 1D and 2D fast spline
# interpolation algorithm (Habermann and Kindermann 2007) in Python.
# Copyright (C) 2012, 2013 Joon H. Ro
# Modified and ported to Micropython by Matthias Kesenheimer, 2025.

# This file is part of Fast-Cubic-Spline-Python.

# Fast-Cubic-Spline-Python is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# Fast-Cubic-Spline-Python is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""
Cubic spline interpolation using Habermann and Kindermann (2007)'s algorithm 
"""

try:
    import platform
    if platform.system() == 'Darwin' or platform.system() == 'Linux':
        from findus.firmware.decorators import micropython
except Exception as _:
    pass

class Spline():
    @staticmethod
    @micropython.native
    def calc_grid(a, b, n:int):
        n = int(n)
        h = 1
        if n > 0:
            h = (b - a) / n
        grid = [0] * int(n + 1)
        for i in range(n + 1):
            grid[i] = i * h + a
        return grid

    @staticmethod
    def interpolate_and_plot(xpoints:list[int], ypoints:list[int]) -> list[int]:
        from matplotlib import pyplot as plt
        a = xpoints[0]
        b = xpoints[-1]
        grid_hat = Spline.calc_grid(a, b, b - a)
        pulse = Spline.pchip_interpolate(xpoints, ypoints, grid_hat)
        plt.clf()
        line_approx = plt.plot(grid_hat, pulse, '-', label='pulse')
        plt.pause(0.001)
        plt.setp(line_approx, linewidth=2, linestyle='-')
        plt.legend()
        plt.show(block=False)
        return pulse

    @staticmethod
    @micropython.native
    def interpolate(x, y, grid_hat):
        n = len(x)
        h = [x[i+1] - x[i] for i in range(n-1)]

        # Solve for the coefficients of the spline:
        # Calculate the system of equations for second derivatives (M)
        alpha = [0] * (n-1)
        for i in range(1, n-1):
            if h[i] != 0 and h[i-1] != 0:
                alpha[i] = (3/h[i]) * (y[i+1] - y[i]) - (3/h[i-1]) * (y[i] - y[i-1])

        l = [1] + [0] * (n-1)
        mu = [0] * (n-1)
        z = [0] * n

        for i in range(1, n-1):
            l[i] = 2 * (x[i+1] - x[i-1]) - h[i-1] * mu[i-1]
            if l[i] != 0:
                mu[i] = h[i] / l[i]
                z[i] = (alpha[i] - h[i-1] * z[i-1]) / l[i]

        l[n-1] = 1
        z[n-1] = 0
        c = [0] * n
        b = [0] * (n-1)
        d = [0] * (n-1)
        a = [y[i] for i in range(n-1)]

        for j in range(n-2, -1, -1):
            c[j] = z[j] - mu[j] * c[j+1]
            if h[i] != 0 and h[j] != 0:
                b[j] = (y[j+1] - y[j]) / h[j] - h[j] * (c[j+1] + 2 * c[j]) / 3
                d[j] = (c[j+1] - c[j]) / (3 * h[j])

        # Interpolate at the query points
        def spline(x_val):
            for i in range(n-1):
                if x[i] <= x_val <= x[i+1]:
                    dx = x_val - x[i]
                    return a[i] + b[i] * dx + c[i] * dx**2 + d[i] * dx**3
            raise ValueError("Query point out of range")

        return [spline(q) for q in grid_hat]

    @staticmethod
    @micropython.native
    def pchip_interpolate(x, y, grid_hat):
        n = len(x)
        h = [x[i+1] - x[i] for i in range(n-1)]  # Interval lengths
        delta = [((y[i+1] - y[i]) / h[i]) if h[i] != 0 else 0 for i in range(n-1)]  # Slopes of secant lines

        # Compute the slopes (m_i)
        m = [0] * n
        m[0] = delta[0]  # Endpoint slope
        m[-1] = delta[-1]  # Endpoint slope

        for i in range(1, n-1):
            if delta[i-1] * delta[i] > 0:  # Monotonicity-preserving condition
                m[i] = 2 / (1/delta[i-1] + 1/delta[i])
            else:
                m[i] = 0  # Flat slope if signs differ

        # Interpolation function
        def hermite(x_val, i):
            t = 0
            if h[i] != 0:
                t = (x_val - x[i]) / h[i]
            h_i = h[i]
            t2 = t * t
            t3 = t2 * t

            h00 = (1 + 2*t) * (1 - t)**2
            h10 = t * (1 - t)**2
            h01 = t2 * (3 - 2*t)
            h11 = t3 - t2

            return (h00 * y[i] +
                    h10 * h_i * m[i] +
                    h01 * y[i+1] +
                    h11 * h_i * m[i+1])

        # Evaluate the query points
        def pchip_eval(x_val):
            for i in range(n-1):
                if x[i] <= x_val <= x[i+1]:
                    return hermite(x_val, i)
            raise ValueError("Query point out of range")

        return [pchip_eval(q) for q in grid_hat]

def pulse_test(i):
    xpoints = [0,   100, 200, 300, 400, 500, 515]
    ypoints = [3.0 + i, 2.0, 2.0, 2.0, 2.0, 0.0, 3.0]
    offset = 3.0
    points_per_volt = 1365
    points_per_ns = 0.1
    tpoints = [0] * len(xpoints)
    vpoints = [0] * len(xpoints)
    for i in range(len(xpoints)):
        tpoints[i] = int(xpoints[i] * points_per_ns)
        vpoints[i] = int((ypoints[i] - offset) * points_per_volt)
    pulse = Spline.interpolate_and_plot(tpoints, vpoints)
    print(pulse)

def pchip_test():
    x = [0, 1, 2, 3, 4, 5, 6]
    y = [0, 1, 0, 1, 1, 1, 1]
    a = x[0]
    b = x[-1]
    grid_hat = Spline.calc_grid(a, b, 50)

    fhat = Spline.pchip_interpolate(x, y, grid_hat)
    print(fhat)
    # plot
    from matplotlib import pyplot as plt
    line_approx = plt.plot(grid_hat, fhat, '-', label='interpolated')
    plt.setp(line_approx, linewidth=2, linestyle='-.')
    plt.legend()
    plt.show()

if __name__ == '__main__':
    #for i in range(200):
    #    pulse_test(i)
    #pulse_test(0)
    #input("Press enter to continue.")
    pchip_test()